from typing import Callable, Dict, List, Optional
import numpy as np
from dp import retrains_iterative
from utils import add_retraining_cost, compute_strategy_cost, vary_cost,strategy_cost_array,compute_metric
from itertools import repeat
from scipy.optimize import fsolve
from pathlib import Path
import ray
import pandas as pd
from pathlib import Path
import re
proj_root = Path(__file__).parent.parent.resolve()
from algorithms import never_retrain,oracle
def find_max_retrain_cost(C: np.ndarray) -> int:
    """Finds the largest retrain cost to study.
        The largest value at which DP solution has same cost as
        never retrain.

    Args:
        C (np.ndarray): The upper-triangular cost matrix

    Returns:
        int: the largest retrain cost rounded up
    """

    def obj(retrain_cost: int, *args) -> float:
        """An objective function comparing cost of optimal strategy to never retrain

        Args:
            retrain_cost (_type_): the retrain cost to evaluate at

        Returns:
            float: the difference between the optimal strategy and the never retrain strategy
        """
        # take the cost matrix for args
        C = args[0]
        # add the retrain cost
        _C = add_retraining_cost(C, retrain_cost)
        # never retrain strategy
        never_retrain_result = never_retrain.run(C, retrain_cost)
        never_retrain_cost = compute_strategy_cost(never_retrain_result, _C)
        # oracle retrain strategy
        oracle_strat = oracle.run(_C, retrain_cost=0)
        oracle_cost = compute_strategy_cost(oracle_strat, _C)
        return never_retrain_cost - oracle_cost

    # find solution using optimization
    sol = fsolve(obj, x0=0, args=(C,))[0]
    return int(sol) + 1

def gather_retrains(C,algorithm,retrain_cost):
    _result = algorithm.optimize(C,retrain_cost)
    _result["algorithm"]=algorithm.name
    _result["retrain_cost"]=retrain_cost
    return _result

def offline_optimize(
    C: np.ndarray,
    algorithms: List,
    max_retrain_cost:int,
    use_ray:bool = False,
):
    # create 100 values in between
    retrain_costs = np.linspace(0, max_retrain_cost, 100)

    results = []
    if use_ray:
        fn = ray.remote(gather_retrains)
        _C = ray.put(C)
    else:
        _C = C        
        fn = gather_retrains
    for cost in retrain_costs:
        for algorithm in algorithms:
            if use_ray:
                _result = fn.remote(_C,algorithm,cost)
            else:
                _result = fn(_C,algorithm,cost)
            results.append(_result)
    
    if use_ray:
        temp = []
        unfinished = results
        while unfinished:
            finished, unfinished = ray.wait(unfinished)
            _result = ray.get(finished[0])
            temp.append(_result)
        results = temp
    return pd.DataFrame.from_dict(results)


def run_online(C,algorithm,retrain_cost,parameters):
    _result = algorithm.run(add_retraining_cost(C,retrain_cost),**parameters)
    _result["retrain_cost"] = retrain_cost
    _result["algorithm"] = algorithm.name
    return _result

def online_execute(
    C: np.ndarray,
    algorithms: List[Callable],
    offline_df: Optional[pd.DataFrame]=None,
    use_ray:bool = False,
):
    # grab the retrain costs to analyse from the offline analysis
    retrain_costs = offline_df.retrain_cost.unique()
    # get the strategy parameters for each strategy as a dictionary 
    
    _df = offline_df.set_index(["retrain_cost","algorithm"])
    results = []
    if use_ray:
        fn = ray.remote(run_online)
        _C = ray.put(C)
    else:
        _C = C        
        fn = run_online
    for cost in retrain_costs:
        for algorithm in algorithms:
            # get the parameters learnt during offline phase
            parameters = _df.loc[(cost, algorithm.name)].parameters
            if use_ray:
                _result = fn.remote(_C,algorithm,cost,parameters)
            else:
                _result = fn(_C,algorithm,cost,parameters)
            results.append(_result)
    if use_ray:
        temp = []
        unfinished = results
        while unfinished:
            finished, unfinished = ray.wait(unfinished)
            _result = ray.get(finished[0])
            temp.append(_result)
        results = temp
    return pd.DataFrame.from_dict(results)


def join_dfs(offline_df,online_df):
    return offline_df.set_index(["algorithm","retrain_cost"]).join(online_df.set_index(["algorithm","retrain_cost"]),rsuffix="_online",lsuffix="_offline").reset_index()

def save_dfs(df:pd.DataFrame,results_dir:Path):
    if not results_dir.exists():
        results_dir.mkdir(exist_ok=True,parents=True)
    algorithms = df.algorithm.unique()
    for algorithm in algorithms:
        df[df.algorithm==algorithm].to_json(results_dir/f"{algorithm}.json",orient="records", lines=True)

def explode_dfs(df:pd.DataFrame):
    # explode the parameters dicts and join back
    return df.join(pd.DataFrame(df["parameters_offline"].to_list()), rsuffix="_param")


def load_dfs(results_dir:Path,verbose=False):
    _dfs = []
    for file in results_dir.glob("**/*.json"):
        if verbose:
            print(f"Loading {file=}")
        search = re.search("model_seed_(\d+)",str(file))
        if search is None:
            model_seed = 0
        else:
            model_seed = search.group(1)
        _temp = pd.read_json(file,lines=True)
        _temp["model_seed"]=model_seed
        _dfs.append(_temp)
    _df = pd.concat(_dfs)
    return _df.reset_index(drop=True)