import numpy as np
import ray
from utils import import_obj
from qdm import compute_data_accuracy, compute_query_accuracy
from omegaconf import DictConfig
from time import perf_counter as time
name = "Adaptive Incremental"

def run(cfg: DictConfig, datas, queries, **kwargs):
    training_time = 0
    model_class = import_obj(cfg.model.adaptive_incremental.class_fn)
    # Get data used to train model
    # Check to make sure the model is incremental
    _partial_fit = getattr(model_class, "partial_fit", None)
    if _partial_fit is None or not callable(_partial_fit):
        raise NotImplementedError(
            f"Partial Fit method not available for {model_class}. Cannot run incremental baseline"
        )
    # get the first batch of data
    _model_data = datas[0]
    if isinstance(_model_data, ray._raylet.ObjectRef):
        _model_data = ray.get(_model_data)
    # train the initial model
    _X = _model_data["X_train"]
    _y = _model_data["y_train"]

    _model = model_class(**cfg.model.adaptive_incremental.kwargs).fit(_X, _y)

    cost_fn = import_obj(cfg.dataset.cost.fn)
    cost_values = {
        "staleness_cost": [],
        "prequential_acc": [],
        "prequential_query_acc": [],
    }
    for t in range(len(datas)):
        # print(f"{t=}")
        data = datas[t]
        if isinstance(data, ray._raylet.ObjectRef):
            data = ray.get(data)

        query = queries[t]
        if isinstance(query, ray._raylet.ObjectRef):
            query = ray.get(query)

        _X = data["X_train"]
        _y = data["y_train"]

        # First compute prequential values
        cost_values["prequential_acc"].append(
            compute_data_accuracy(
                model=_model, model_data=_model_data, new_data=data, query_data=query
            )
        )
        cost_values["prequential_query_acc"].append(
            compute_query_accuracy(
                model=_model, model_data=_model_data, new_data=data, query_data=query
            )
        )
        
        if t>0:
            start = time()
            # Incrementally update the model with the batch data
            _model = _model.partial_fit(_X, _y)
            training_time += time() - start
            # update the model's data to the most recent batch
            # _model_data["X_train"] = np.append(_model_data["X_train"],_X,axis=0)
            # _model_data["y_train"] = np.append(_model_data["y_train"],_y)
            _model_data["X_train"] = _X
            _model_data["y_train"] = _y

        # compute cost function
        cost_values["staleness_cost"].append(
            cost_fn(
                model=_model, model_data=_model_data, new_data=data, query_data=query
            )
        )

    # convert costs into arrays
    cost_values = {name: np.array(costs) for name, costs in cost_values.items()}
    _retrains = list(range(len(datas)))
    running_time = {
        "total":training_time,
        "training_time":training_time
    }
    result = {
        "retrains": _retrains,
        "num_retrains": len(_retrains),
        "parameters": {},
        "cost_fns_values": cost_values,
        "num_partial_fits": len(datas) - 1,
        "running_time":running_time
    }
    return result
