import numpy as np
from skmultiflow.drift_detection.adwin import ADWIN
import ray
from utils import import_obj
from copy import deepcopy
from qdm import compute_data_accuracy, compute_query_accuracy
from time import perf_counter as time
name = "ADWIN+Incremental"


def run(cfg, datas, queries, **kwargs):
    training_time = 0
    detector_time = 0

    model_class = import_obj(cfg.model.class_fn)
    # Get data used to train model
    # Check to make sure the model is incremental
    _partial_fit = getattr(model_class, "partial_fit", None)
    if _partial_fit is None or not callable(_partial_fit):
        raise NotImplementedError(
            f"Partial Fit method not available for {model_class}. Cannot run incremental baseline"
        )
    # get the first batch of data
    _model_data = datas[0]
    if isinstance(_model_data, ray._raylet.ObjectRef):
        _model_data = ray.get(_model_data)
    # train the initial model
    _X = _model_data["X_train"]
    _y = _model_data["y_train"]
    _model = model_class(**cfg.model.kwargs).fit(_X, _y)

    cost_fn = import_obj(cfg.dataset.cost.fn)
    cost_values = {
        "staleness_cost": [],
        "prequential_acc": [],
        "prequential_query_acc": [],
    }
    # init the drift detector
    start = time()
    detector = ADWIN()
    detector_time += time() - start

    update = False
    _updates = set()
    for t in range(len(datas)):
        data = datas[t]
        if isinstance(data, ray._raylet.ObjectRef):
            data = ray.get(data)

        if isinstance(_model, ray._raylet.ObjectRef):
            _model = ray.get(_model)

        query = queries[t]
        if isinstance(query, ray._raylet.ObjectRef):
            query = ray.get(query)

        _X = data["X_train"]
        _y = data["y_train"]

        # First compute prequential values
        cost_values["prequential_acc"].append(
            compute_data_accuracy(
                model=_model, model_data=_model_data, new_data=data, query_data=query
            )
        )
        cost_values["prequential_query_acc"].append(
            compute_query_accuracy(
                model=_model, model_data=_model_data, new_data=data, query_data=query
            )
        )

        # predict for detector
        start = time()
        _yhat = _model.predict(_X)
        errors = _yhat != _y

        for i, _e in enumerate(errors):
            detector.add_element(_e)
            # if any drift is detected in the current batch update
            if detector.detected_change():
                _updates.add(t)
                update = True
        detector_time += time() - start
        # do not update in first batch
        if update and t>0:
            start = time()
            _model = _model.partial_fit(_X, _y)
            training_time += time() - start
            _model_data["X_train"] = _X
            _model_data["y_train"] = _y
            update = False

        # Compute staleness cost after decision has been made
        cost_values["staleness_cost"].append(
            cost_fn(
                model=_model, model_data=_model_data, new_data=data, query_data=query
            )
        )

    # convert costs into arrays
    cost_values = {name: np.array(costs) for name, costs in cost_values.items()}
    _updates = list(_updates)
    running_time = {
        "total":training_time+detector_time,
        "training_time":training_time,
        "detector_time":detector_time
    }
    result = {
        "retrains": _updates,
        "num_retrains": len(_updates),
        "parameters": {},
        "cost_fns_values": cost_values,
        "partial_fits": _updates, 
        "num_partial_fits": len(_updates),
        "running_time":running_time
    }
    return result
