import argparse
import os
from html import parser
from pathlib import Path

import numpy as np


proj_root = Path(__file__).parent.parent.resolve()
import logging
import shutil
from importlib import import_module
from time import perf_counter as time

import hydra
import pandas as pd
import ray
from hydra.core.hydra_config import HydraConfig
from hydra.utils import get_original_cwd
from omegaconf import DictConfig, OmegaConf
import joblib

from analyse import join_dfs, offline_optimize, online_execute, save_dfs
from baseline import gather_baselines
from qdm import compute_data_accuracy, compute_query_accuracy
from utils import (_compute_prequential_accuracy, _compute_strategy_cost,
                   compute_cost_matrix_multiple, get_model_dict, import_obj,
                   save_matrices,save_model_dicts)

log = logging.getLogger(__name__)


def train_models(cfg:DictConfig,offline_datas,online_datas):
    model_class = import_obj(cfg.model.class_fn)
    log.info(f"Training {cfg.model.name} models")
    offline_models = get_model_dict(
        offline_datas, model_class, use_ray=cfg.use_ray, **cfg.model.kwargs
    )
    online_models = get_model_dict(online_datas, model_class, use_ray=cfg.use_ray, **cfg.model.kwargs)
    return offline_models,online_models

@hydra.main(version_base=None,config_path="config")
def run(cfg):
    load_fn = import_obj(cfg.dataset.load_fn)
    log.info(f"Creating Cost Matrices {cfg.dataset.name} dataset")
    offline_datas, online_datas, offline_queries, online_queries = load_fn(cfg)
    log.info("Gathered data and queries")
    results_dir:Path = proj_root / cfg.results_dir
    # Train Models
    if "random_state" in cfg.model.kwargs:
       model_seed = cfg.model.kwargs.random_state
       seed_results_dir:Path = results_dir/f"model_seed_{model_seed}"
    else:
        model_seed = np.nan
        seed_results_dir = results_dir
    start = time()
    offline_models,online_models = train_models(cfg,offline_datas,online_datas)
    elapsed = time() - start
    log.info(f"Trained Offline and Online Models with seed: {model_seed} in {elapsed:g}sec")
    
    # compute offline cost matrix
    cost_fn = import_obj(cfg.dataset.cost.fn)
    cost_fns = [cost_fn,compute_data_accuracy,compute_query_accuracy]
    cost_fn_kwargs = dict(cfg.dataset.cost.kwargs)
    # if a custom qdm_fn was defined load it 
    if "qdm_fn" in cost_fn_kwargs:
        cost_fn_kwargs["qdm_fn"] = import_obj(cost_fn_kwargs["qdm_fn"])
    # put all kwargs for cost functions in one place
    cost_kwargs =[cost_fn_kwargs,{},{}]

    start = time()
    offline_C,offline_data_acc,offline_query_acc = compute_cost_matrix_multiple(
        offline_datas, offline_models, cost_fns, offline_queries, use_ray=cfg.use_ray, cost_kwargs=cost_kwargs
    )
    elapsed = time() - start
    log.info(f"Computed {len(cost_fns) } offline Matrices in {elapsed:g}sec")
    # compute online cost matrix
    start = time()
    
    online_C,online_data_acc,online_query_acc = compute_cost_matrix_multiple(
        online_datas, online_models, cost_fns, online_queries, use_ray=cfg.use_ray, cost_kwargs=cost_kwargs
    )
    elapsed = time() - start
    log.info(f"Computed {len(cost_fns)} online Matrices in {elapsed:g}sec")
    
    if not seed_results_dir.exists():
        seed_results_dir.mkdir(parents=True,exist_ok=True)
    if cfg.save_costs:
        log.info("Saving Matrices to numpy compressed format")
        save_matrices({
                "offline_C":offline_C,
                "offline_data_acc":offline_data_acc,
                "offline_query_acc":offline_query_acc,
                "online_C":online_C,
                "online_data_acc":online_data_acc,
                "online_query_acc":online_query_acc,
            },
            seed_results_dir)
    if cfg.save_models:
        log.info("Saving Models to joblib")
        save_model_dicts(offline_models,online_models,seed_results_dir,use_ray=cfg.use_ray)
    
    log.info("Logging Configs and Logs to results directorty")
    log_dir = seed_results_dir/"log"
    hydra_log_dir = HydraConfig.get().runtime.output_dir
    shutil.copytree(hydra_log_dir,str(log_dir),dirs_exist_ok=True)



if __name__ == "__main__":
    run()
