from pathlib import Path

import numpy as np
import pandas as pd
from omegaconf import DictConfig

proj_root = Path(__file__).parent.parent.parent.resolve()
import scipy.sparse as sparse
from sklearn.preprocessing import MaxAbsScaler, OneHotEncoder
from utils import get_online_offline_data_dict,get_online_offline_data_dict_with_onehot
from typing import List

def get_data(
    data_dir: Path,
    categorical_features: List[str] = [
        "Airline",
        "Flight",
        "AirportFrom",
        "AirportTo",
        "DayOfWeek",
    ],
    numerical_features: List[str] = [
        "Time",
        "Length"
    ],
    T: int = 100,
    onehot=True,
):
    airlines_dir = data_dir / "airlines"
    df = pd.read_csv(airlines_dir / "raw.csv")
    if onehot:
        X_categorical = df.loc[:, categorical_features].values
        X_numerical = df.loc[:, numerical_features].values
        y = df.iloc[:, -1].values
        enc = OneHotEncoder()
        X_one_hot = enc.fit_transform(X_categorical)

        X_full = sparse.hstack([X_one_hot, X_numerical])
        scaler = MaxAbsScaler()
        X_scaled = scaler.fit_transform(X_full).tocsr()
        X_split = []
        y_split = []
        N = X_full.shape[0]
        num_batches = T
        points_per_batch = int(N / 100) + 1
        for i in range(num_batches):
            X_split.append(
                X_scaled[i * points_per_batch : (i + 1) * points_per_batch].tocoo()
            )
            y_split.append(y[i * points_per_batch : (i + 1) * points_per_batch])
    else:
        X_full = df.iloc[:,:-1].values
        y = df.iloc[:,-1].values
        X_split = []
        y_split = []
        N = X_full.shape[0]
        num_batches = T
        points_per_batch = int(N / 100) + 1
        for i in range(num_batches):
            X_split.append(
                X_full[i * points_per_batch : (i + 1) * points_per_batch]
            )
            y_split.append(y[i * points_per_batch : (i + 1) * points_per_batch])
    return X_split, y_split


def load_dicts(cfg: DictConfig):
    data_dir = proj_root / cfg.dataset.data.dir
    X_full, y_full = get_data(
        data_dir=data_dir, 
        categorical_features=cfg.dataset.categorical_features,
        numerical_features= cfg.dataset.numerical_features,
        T=cfg.dataset.num_batches, 
        onehot=cfg.dataset.onehot,
    )
    offline_datas, online_datas = get_online_offline_data_dict(
        X_full,
        y_full,
        offline_split=cfg.dataset.offline_split,
        use_ray=cfg.use_ray,
        queries_split=cfg.dataset.queries.queries_split,
    )
    if cfg.dataset.queries.name == "data":
        return offline_datas, online_datas, offline_datas, online_datas

def load_dicts_with_onehot(cfg:DictConfig):
    # When using categorical features as is include the one-hot features for QDM computation
    data_dir = proj_root / cfg.dataset.data.dir
    X_full, y_full = get_data(
        data_dir=data_dir, 
        categorical_features=cfg.dataset.categorical_features,
        numerical_features= cfg.dataset.numerical_features,
        T=cfg.dataset.num_batches, 
        onehot=cfg.dataset.onehot,
    )
    X_onehot,_ = get_data(
        data_dir=data_dir, 
        categorical_features=cfg.dataset.categorical_features,
        numerical_features= cfg.dataset.numerical_features,
        T=cfg.dataset.num_batches, 
        onehot=True,
    )
    offline_datas, online_datas = get_online_offline_data_dict_with_onehot(
        X_full,
        X_onehot,
        y_full,
        offline_split=cfg.dataset.offline_split,
        use_ray=cfg.use_ray,
        queries_split=cfg.dataset.queries.queries_split,
    )
    if cfg.dataset.queries.name == "data":
        return offline_datas, online_datas, offline_datas, online_datas