import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from pandas import concat
from sklearn.datasets import make_blobs

project_root = Path(__file__).parent.parent.parent.resolve()
sys.path.append(str((project_root/"src").resolve()))
import ray
from utils import get_online_offline_data_dict, get_online_offline_query_dict, generate_static_queries


def generate_data(num_batches:int=100, points_per_batch:int=1000, seed:int=0, alpha:float=1.0,std:float=0.03,cov_offset:Optional[float]=0.0):
    rng = np.random.default_rng(seed=seed)
    X = []
    for t in range(num_batches):
        # Covariate shift of the region of data
        loc = ((t + 1) % 7) / 10 + cov_offset
        _X = rng.normal(loc=loc, scale=std, size=(points_per_batch, 2))
        X.append(_X)
    y = classify_data(X,alpha=alpha)
    return X,y

def classify_data(X:List[np.ndarray],alpha:float=0.01,**kwargs):
    drift = False
    y = []
    for t,_X in enumerate(X):
        # every 10 batches switch inequality for concept drift
        if t%10 == 0:
            drift = ~drift
        # decide on labels based on drift 
        if not drift:
            _y = (alpha * np.sin(np.pi * _X[:, 0]) > _X[:, 1]).astype(int)
        else:
            _y = (alpha * np.sin(np.pi * _X[:, 0]) < _X[:, 1]).astype(int)
        y.append(_y)
    return y

def plot_data(data_dict:Dict[int,dict],query_dict:Optional[Dict[int,dict]]=None,alpha:float=1,**kwargs):
    """Plots the data assuming a covcon concept

    Args:
        data_dict (Dict[int,dict]): the data dictionary. keys are batches and values are dicts.
        query_dict (Optional[Dict[int,dict]], optional): Query dictionary to plot query points. If none uses samples from data as queries. Defaults to None.
        alpha (float, optional): the value of alpha for the covcon function. Defaults to 1.
    """
    
    T = len(data_dict)
    num_rows = ((T+1)//4)+1
    _,ax = plt.subplots(num_rows,4,figsize=(5*4,5*num_rows))
    ax = ax.flatten()
    x = np.linspace(-0.5,1)
    line = alpha*np.sin(np.pi*x)
    colors= {0:"yellow",1:"purple"}
    color_map = np.vectorize(colors.get)
    for t in range(T):
        data = data_dict[t]
        if isinstance(data,ray._raylet.ObjectRef):
            data = ray.get(data)
        X = data["X_train"]
        y = data["y_train"]
        ax[t].scatter(X[:,0],X[:,1],c=color_map(y))
        ax[t].plot(x,line,"b--")
        ax[t].set_xlim(-0.5,1)
        ax[t].set_ylim(-0.5,1)
        if query_dict is not None:
            query =query_dict[t]
            if isinstance(query,ray._raylet.ObjectRef):
                query = ray.get(query)
            X_q = query["X_query"]
        else:
            X_q = data["X_query"]
        ax[t].scatter(X_q[:,0],X_q[:,1],ec="red",c=None)
        ax[t].set_title(f"{t=}")
    plt.show()



def load_dicts(cfg):
    X_full, y_full = generate_data(**cfg.dataset.data.kwargs)
    offline_datas, online_datas = get_online_offline_data_dict(
        X_full,
        y_full,
        offline_split=cfg.dataset.offline_split,
        use_ray=cfg.use_ray,
        queries_split = cfg.dataset.queries.queries_split
    )
    if cfg.dataset.queries.name == "data":
        return offline_datas, online_datas, offline_datas, online_datas
    elif "static" in cfg.dataset.queries.name:
        X_queries = generate_static_queries(**cfg.dataset.queries.kwargs)
        # classify to get ground truth query labels
        y_queries = classify_data(X_queries,**cfg.dataset.data.kwargs)
        offline_queries,online_queries = get_online_offline_query_dict(X_queries,y_queries,offline_split=cfg.dataset.offline_split,use_ray=cfg.use_ray)
    else:
        X_queries, y_queries = generate_data(**cfg.dataset.queries.kwargs)
        offline_queries, online_queries = get_online_offline_query_dict(
            X_queries, y_queries, cfg.dataset.offline_split, use_ray=cfg.use_ray
        )
    return offline_datas, online_datas, offline_queries, online_queries
