from typing import Dict, Optional

import matplotlib.pyplot as plt
import numpy as np
from sklearn.datasets import make_blobs
from sklearn.linear_model import SGDClassifier
from sklearn.preprocessing import MinMaxScaler


def plot_model_data(model,data,ax=None):
    if ax is None:
        fig,ax = plt.subplots()
    
    Xd,Xq,yd,yq = data.values()
    X = np.r_[Xd,Xq]
    y = np.r_[yd,yq]
    w = model.coef_[0]
    b = model.intercept_
    # x = np.linspace(X[:,0].min(),X[:,0].max())
    x = np.linspace(-20,20)
    line = -(b+w[0]*x)/w[1]
    ax.plot(x,line,"b--",label="model")
    # plt.ylim(X[:,1].min()-1,X[:,1].max()+1)
    ax.set_ylim(-10,10)
    _plot_data(data,ax)
    return ax
    # plt.legend()
    # plt.show()

def _plot_data(data,ax=None,alpha=1):
    if ax is None:
        fig,ax = plt.subplots()
    
    Xd,Xq,yd,yq = data.values()
    X = np.r_[Xd,Xq]
    y = np.r_[yd,yq]
    ax.scatter(Xd[:,0],Xd[:,1],c=yd,label="data",alpha=alpha)
    ax.scatter(Xq[:,0],Xq[:,1],c=yq,edgecolors="red",label="queries",alpha=alpha)
    ax.set_xlim(-20,20)
    ax.set_ylim(-20,20)
    return ax
    # plt.legend()
    # plt.show()


def translate_centres(t,centres,move=8,**kwargs):
    c1,c2 = centres
    return np.array([c1 + t*move, c2 +t*move])

def roundtrip_centres(t,centres,move=8,**kwargs):
    c1,c2 = centres
    if t<0.5:
        return np.array([c1 + t*2*move, c2 +t*2*move])
    else:
        return np.array([c1+(1-t)*2*move,c2+(1-t)*2*move])

def rotate_centres(t,centres,rad=2.5,*kwargs):
    c1,c2 = centres
    rot_matrix = np.array([[np.cos(t*rad),-np.sin(t*rad)],[np.sin(t*rad),np.cos(t*rad)]])

    return centres@rot_matrix

def collide_centres(t,centers,**kwargs):
    c1,c2 = centers
    return np.array([t*(c1)+(1-t)*c2,(1-t)*c1 + t*c2])

def get_centres(t,centers, change_method=rotate_centres, **kwargs):
    return change_method(t,centers,**kwargs)

def get_samples(n_samples,centers):
    X,y = make_blobs(n_samples=n_samples,n_features=2, centers=centers) 
    return X,y

def get_data(t,arrival_rv,init_centres,data_prob:float, **kwargs):
    centers = get_centres(t,init_centres,**kwargs)
    n_samples = arrival_rv.rvs()
    n_data = int(n_samples*data_prob)
    n_queries = n_samples - n_data
    # print(n_samples)
    # X,y = make_blobs(n_samples=n_samples,n_features=2, centers=centers) 
    X_data,y_data = get_samples(n_data,centers)
    if 'query_centers_shift' in kwargs:
        c1,c2=centers
        t2,t1 = 0-kwargs['query_centers_shift'],1+kwargs['query_centers_shift']
        q_centers = np.array([t1*c1+(1-t1)*c2,t2*c1+(1-t2)*c2])
        X_query,y_query = get_samples(n_queries,q_centers)
    else:
        X_query,y_query = get_samples(n_queries,centers)

    # data = train_test_split(X,y,train_size=data_prob)
    return X_data,X_query,y_data,y_query

def get_data_query(T:int,arrival_rv,init_centres,data_prob:float,**kwargs):
    t_s = np.linspace(0,1,T+1) # skip t=0 as that is init data
    X_data,y_data,X_query,y_query = [],[],[],[]
    for i,t in enumerate(t_s):
        data = get_data(t,arrival_rv,init_centres,data_prob,**kwargs)
        X_data.append(data[0])
        X_query.append(data[1])
        y_data.append(data[2])
        y_query.append(data[3])
    scaler = MinMaxScaler()
    X_data = np.concatenate(X_data)
    y_data = np.concatenate(y_data)
    X_query = np.concatenate(X_query)
    y_query = np.concatenate(y_query)

    scaler.fit(np.concatenate([X_data,X_query]))

    X_data = scaler.transform(X_data)
    X_query = scaler.transform(X_query) 
    X_data_split = np.array_split(X_data,T+1)
    y_data_split = np.array_split(y_data,T+1)
    X_query_split = np.array_split(X_query,T+1)
    y_query_split = np.array_split(y_query,T+1)
    return X_data_split,y_data_split,X_query_split,y_query_split

def update_model(model,X,y):
    # return model.fit(X,y)
    return model.partial_fit(X,y)

def get_metrics(model:SGDClassifier,data,inverse=False):
    # data_pred_prob = model.predict_proba(data["X_train"])[:,1]
    # query_pred_prob = model.predict_proba(data["X_query"])[:,1]
    # try:
    #     roc_data = roc_auc_score(data["y_train"],data_pred_prob)
    #     roc_query = roc_auc_score(data["y_query"],query_pred_prob)
    # except Exception as e:
    #     roc_data = 0
    #     roc_query = 0
    data_preds = model.predict(data["X_train"])
    query_preds = model.predict(data["X_query"])
    if not inverse:
        data_correct = (data_preds==data["y_train"]).sum()
        query_correct = (query_preds == data["y_query"]).sum()
        return data_correct,query_correct
    else:
        data_incorrect = (data_preds!=data["y_train"]).sum()
        query_incorrect = (query_preds != data["y_query"]).sum()
        return data_incorrect,query_incorrect

def plot_data(data_dict:Dict[int,dict],query_dict:Optional[Dict[int,dict]]=None):
    """Plots the data assuming 

    Args:
        data_dict (Dict[int,dict]): the data dictionary. keys are batches and values are dicts.
        query_dict (Optional[Dict[int,dict]], optional): Query dictionary to plot query points. If none uses samples from data as queries. Defaults to None.
        alpha (float, optional): the value of alpha for the covcon function. Defaults to 1.
    """
    
    T = len(data_dict)
    num_rows = ((T+1)//4)+1
    _,ax = plt.subplots(num_rows,4,figsize=(5*4,5*num_rows))
    ax = ax.flatten()
    colors= {0:"yellow",1:"purple"}
    color_map = np.vectorize(colors.get)
    for t in range(T):
        data = data_dict[t]
        X = data["X_train"]
        y = data["y_train"]
        ax[t].scatter(X[:,0],X[:,1],c=color_map(y))
        ax[t].set_xlim(-0.5,1)
        ax[t].set_ylim(-0.5,1)
        if query_dict is not None:
            X_q = query_dict[t]["X_query"]
        else:
            X_q = data["X_query"]
        ax[t].scatter(X_q[:,0],X_q[:,1],ec="red",c=None)
        ax[t].set_title(f"{t=}")
    plt.show()
