# %%
from IPython import get_ipython

if get_ipython() is not None and __name__ == "__main__":
    notebook = True
    get_ipython().run_line_magic("load_ext", "autoreload")
    get_ipython().run_line_magic("autoreload", "2")
else:
    notebook = False
import sys
from pathlib import Path

project_root = Path(__file__).resolve().parent.parent
sys.path.append(str((project_root / "src").resolve()))
from bisect import bisect_left
from collections import OrderedDict

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from analyse import load_dfs
from dp import add_retraining_cost
from utils import load_cost_matrices, plot_retrains, strategy_cost_array,_plot_single_retrain

# %%

columnwidth = 234.8775
fullwidth = 487.8225

label_rename = OrderedDict({
    "Oracle": r"\textsc{Oracle}",
    "Periodic": r"\textsc{Cara-P}",
    "Threshold": r"\textsc{Cara-T}",
    "Cumulative Threshold": r"\textsc{Cara-CT}",
    "Never Retrain": r"\textsc{NR}",
    "Markov": r"\textsc{Markov}",
    "ADWIN": r"\textsc{ADWIN}",
    "DDM": r"\textsc{DDM}",
    "ADWIN+Incremental": r"\textsc{ADWIN-Update}",
    "Incremental": r"\textsc{Always Update}",
    "Adaptive Incremental": r"\textsc{HAT-Update}",
    "Always Retrain":r"\textsc{Always Retrain}"
})


hue_order = list(label_rename.values())
_default_palette = {}
for k,color in zip(label_rename,sns.color_palette(n_colors=len(label_rename))):
    _default_palette[k]=color
default_palette = {label_rename[k]:v for k,v in _default_palette.items()}
# just set the color for the generic variant same as that of CT
default_palette[r"\textsc{Cara}"] = default_palette[r"\textsc{Cara-CT}"]
default_palette[r"\textsc{Always Retrain}"] = default_palette[r"\textsc{NR}"]
default_palette[r"\textsc{Cara-T-Retrain}"] = default_palette[r"\textsc{Cara-T}"]
default_palette[r"\textsc{ADWIN-Retrain}"] = default_palette[r"\textsc{ADWIN}"]
label_rename["Cara"] = r"\textsc{Cara}"

dataset_query_dists = [
    ("covcon","data"), # 0
    ("covcon","static_middle"), #1
    ("circles","data"), #2
    ("circles","static_middle"), #3
    ("gauss","data"), # 4
    ("gauss","static_middle"), #5
    ("elec","data"), #6
    ("airlines","data"), #7
    ("covertype","data"), #8
    ]
dataset_names = [
    r"\textsc{CovCon-D}",
    r"\textsc{CovCon-S}",
    r"\textsc{Circle-D}",
    r"\textsc{Circle-S}",
    r"\textsc{Gauss-D}",
    r"\textsc{Gauss-S}",
    r"\textsc{Electricity}",
    r"\textsc{Airlines}",
    r"\textsc{Covertype}",
]

DATASET_MAPPING = {
    ("covcon","data"):r"\textsc{CovCon-D}",
    ("covcon","static_middle"):    r"\textsc{CovCon-S}",
    ("circles","data"):    r"\textsc{Circle-D}",
    ("circles","static_middle"):    r"\textsc{Circle-S}",
    ("gauss","data"):    r"\textsc{Gauss-D}",
    ("gauss","static_middle"):    r"\textsc{Gauss-S}",
    ("elec","data"):    r"\textsc{Electricity}",
    ("airlines","data"):    r"\textsc{Airlines}",
    ("covertype","data"):    r"\textsc{Covertype}",
}

# save_fig=False
# figures_dir = project_root / "paper" / "ICML_2023" / "figs"
# tables_dir = project_root / "paper" / "ICML_2023" / "tables"
# figures_dir = project_root / "paper" / "KBS_2024" / "figs"
# tables_dir = project_root / "paper" / "KBS_2024" / "tables"
# results_base_dir = project_root / "results"
# figures_dir = project_root / "paper" / "KBS_2024" / "figs_HT"
# tables_dir = project_root / "paper" / "KBS_2024" / "tables_HT"
# results_base_dir = project_root / "results_HT"
figures_dir = project_root / "paper" / "KBS_revision_2024" 
# tables_dir = project_root / "paper" / "KBS_2024" / "tables_HT"
results_base_dir = project_root / "results_HT"
#%%
def set_size(width, fraction=1, subplots=(1, 1)):
    """Set figure dimensions to avoid scaling in LaTeX.

    Parameters
    ----------
    width: float or string
            Document width in points, or string of predined document type
    fraction: float, optional
            Fraction of the width which you wish the figure to occupy
    subplots: array-like, optional
            The number of rows and columns of subplots.
    Returns
    -------
    fig_dim: tuple
            Dimensions of figure in inches
    """
    if width == 'thesis':
        width_pt = 426.79135
    elif width == 'beamer':
        width_pt = 307.28987
    else:
        width_pt = width

    # Width of figure (in pts)
    fig_width_pt = width_pt * fraction
    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Golden ratio to set aesthetic figure height
    # https://disq.us/p/2940ij3
    golden_ratio = (5**.5 - 1) / 2

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt
    # Figure height in inches
    fig_height_in = fig_width_in * golden_ratio * (subplots[0] / subplots[1])

    return (fig_width_in, fig_height_in)

def load_df(dataset,
            query_dist,
            label_mapping=label_rename,
            result_base_dir=results_base_dir,
            verbose=False
            ):
    result_dir = result_base_dir / dataset / query_dist
    df = load_dfs(result_dir, verbose=verbose)
    if label_mapping is not None:
        df["algorithm"] = df.algorithm.apply(label_rename.get)
    return df

def get_strategy_cost_ylim(df,baselines=["ADWIN","DDM"],online=False):
    baselines_rename = map(label_rename.get,baselines)
    if online:
        return df[~df.algorithm.isin(baselines_rename)].strategy_cost_online.max()
    else:
        return df[~df.algorithm.isin(baselines_rename)].strategy_cost_offline.max()


# %%
def plot_column(df,
                column: str = "strategy_cost",
                online=True,
                ax=None,
                legend=False):
    if ax is None:
        fig, ax = plt.subplots()

    if online:
        suffix = "online"
    else:
        suffix = "offline"

    if f"{column}_{suffix}" not in df.columns:
        raise ValueError(f"Column {column}_{suffix} is not present in {df.columns=}")

    sns.lineplot(
        data=df,
        x="retrain_cost",
        y=f"{column}_{suffix}",
        hue="algorithm",
        style="algorithm",
        #  markers=True,
        markevery=5,
        ax=ax,
        markersize=10,
        # hue_order=hue_order,
        palette=default_palette
        )
    ax.set_xlabel("Retraining Cost $\kappa$")
    if not legend:
        ax.get_legend().remove()
    else:
        ax.legend(bbox_to_anchor=(1, 0.5),title="Algorithms", loc="center left")


def plot_costs_retrains(df,online:bool=True, ax=None):
    if ax is None:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 10))
    elif len(ax) != 2:
        raise ValueError(f"Requires exactly 2 axes, received: {len(ax)} axes")
    else:
        ax1, ax2 = ax
    ylim = get_strategy_cost_ylim(df,online=online)
    plot_column(df,"strategy_cost",online,ax=ax1,legend=False)
    ax1.set_ylim(top=ylim*1.05)
    plot_column(df,"num_retrains",online,ax=ax2,legend=True)

def plot_accs(df,online:bool=True, ax=None):
    if ax is None:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 10))
    elif len(ax) != 2:
        raise ValueError(f"Requires exactly 2 axes, received: {len(ax)} axes")
    else:
        ax1, ax2 = ax
    plot_column(df,"prequential_acc",online,ax=ax1,legend=False)
    ax1.set_ylim(0,1)
    plot_column(df,"prequential_query_acc",online,ax=ax2,legend=True)
    ax2.set_ylim(0,1)


def plot_box_cost(df, hue_order, ax=None):
    if ax is None:
        _, ax = plt.subplots()

    sns.boxplot(data=df,
                x="algorithm",
                y="strategy_cost_online",
                ax=ax,
                order=hue_order)
    ax.tick_params(axis='x', rotation=90)
    ax.set_ylabel("Strategy Cost")


def plot_bar_acc(df, hue_order, ax=None,query=False):
    if ax is None:
        _, ax = plt.subplots()
    if query:
        suffix = "_query"
    else:
        suffix = ""
    sns.barplot(data=df,
                x="algorithm",
                y=f"prequential{suffix}_acc_online",
                ax=ax,
                capsize=0.2,
                order=hue_order)
    ax.tick_params(axis='x', rotation=90)
    if query:
        ax.set_ylabel("Query Accuracy")
    else:
        ax.set_ylabel("Data Accuracy")


def plot_bar_retrains(df, hue_order, online: bool, ax=None):
    if ax is None:
        _, ax = plt.subplots()
    if online:
        suffix = "online"
    else:
        suffix = "offline"
    sns.barplot(data=df,
                x="algorithm",
                y=f"num_retrains_{suffix}",
                ax=ax,
                capsize=0.2,
                order=hue_order)
    ax.set_ylabel(f"Number of retrains ({suffix})")
    ax.tick_params(axis='x', rotation=90)


# %%


def get_cum_costs(retrains, _C):
    try:
        cum_costs = np.cumsum(strategy_cost_array(retrains, _C))
    except Exception as e:
        cum_costs = np.zeros(len(_C))

    return cum_costs


# %%


def get_closest_cost(request: float, retrain_costs: np.ndarray):
    _retrain_costs = retrain_costs.copy()
    _retrain_costs.sort()
    _retrain_costs_ceil = np.floor(_retrain_costs)
    idx = bisect_left(_retrain_costs_ceil, request)
    cost = sorted(_retrain_costs)[idx]
    print(
        f"Retrain Cost closest to input {request} is {cost}. Performing Analysis"
    )
    return cost


def plot_retains_heatmap(df: pd.DataFrame,
                         C: np.ndarray,
                         retrain_cost: float,
                         algorithm: str,
                         online: bool = False,
                         color="r",
                         ax=None,
                         save_fig=False,
                         figure_dir: Path = figures_dir,
                         **kwargs):
    if ax is None:
        _, ax = plt.subplots()
    if online:
        suffix = "online"
    else:
        suffix = "offline"
    _C = add_retraining_cost(C, retrain_cost)
    row = df[(df.algorithm == algorithm)
             & (df.retrain_cost == retrain_cost)].iloc[0]
    retrains = row[f"retrains_{suffix}"]
    im = plot_retrains(_C, [retrains], [color], ax=ax, **kwargs)
    return im


def plot_heatmaps(dataset,
                  query_dist,
                  request_cost: float,
                  results_base_dir=results_base_dir,
                  save_fig: bool = False,
                  figure_dir: Path = figures_dir):
    df = load_df(dataset, query_dist, result_base_dir=results_base_dir)
    results_dir = results_base_dir / dataset / query_dist
    matrices = load_cost_matrices(results_dir / "costs.npz")
    offline_C, online_C = matrices["offline_C"],matrices["online_C"]
    offline_iu = np.triu_indices(len(offline_C))
    online_iu = np.triu_indices(len(online_C))
    vmin, vmax = min(offline_C[offline_iu].min(),
                     online_C[online_iu].min()), max(
                         offline_C[offline_iu].max(),
                         online_C[online_iu].max())
    kwargs = {"vmin": vmin, "vmax": vmax}
    retrain_costs = df.retrain_cost.unique()
    cost = get_closest_cost(request_cost, retrain_costs)
    fig, (ax1, ax2, ax3) = plt.subplots(1,
                                        3,
                                        figsize=(10, 5),
                                        gridspec_kw={"width_ratios":[1, 1, 0.05]})
    plot_retains_heatmap(df,
                         offline_C,
                         cost,
                         "Oracle",
                         online=False,
                         ax=ax1,
                         **kwargs)

    ax1.set_title("Offline Cost Matrix")
    ax1.set_xlabel("Batch evaluated At")
    ax1.set_ylabel("Batch trained At")
    ax1.legend(["Optimal Oracle"], loc="lower left")

    im2 = plot_retains_heatmap(df,
                               online_C,
                               cost,
                               "Oracle",
                               online=True,
                               ax=ax2,
                               **kwargs)
    ax2.set_title("Online Cost Matrix")
    ax2.set_xlabel("Batch evaluated At")
    ax2.set_ylabel("Batch trained At")
    ax2.legend(["Optimal Oracle"], loc="lower left")
    fig.colorbar(im2, cax=ax3, shrink=0.6,label="Cost")
    fig.suptitle(
        f"Cost Matrices and Optimal Oracle Solutions given $\kappa={int(cost):g}$"
    )
    fig.subplots_adjust()
    if save_fig:
        plt.savefig(figure_dir / f"{dataset}_{query_dist}_heatmap.pdf",
                    bbox_inches="tight")
    else:
        plt.show()


# %%
def plot_opt_parameters(dataset,
                        query_dist,
                        request_cost: float,
                        save_fig: bool = False,
                        figure_dir: Path = figures_dir):
    df = load_df(dataset, query_dist, result_base_dir=results_base_dir)
    exploded_df = df.join(pd.DataFrame(df["parameters_offline"].to_list()),
                          rsuffix="_param")
    retrain_costs = df.retrain_cost.unique()
    cost = get_closest_cost(request_cost, retrain_costs)

    fig, ax = plt.subplots(1, 3, figsize=(15, 5))
    items = ["threshold", "cum_threshold", "period"]
    item_labels = [
        "optimal thresholds", "optimal cumulative thresholds",
        "optimal periodicities"
    ]
    ylabels = [r"$\tau$", r"$\tau_{cum}$", r"$\phi$"]
    for i, (item, label, ylabel) in enumerate(zip(items, item_labels,
                                                  ylabels)):
        sns.lineplot(data=exploded_df, x="retrain_cost", y=item, ax=ax[i])
        ax[i].set_title(label)
        ax[i].set_ylabel(ylabel)
        ymax = exploded_df[exploded_df.retrain_cost ==
                           cost][item].dropna().values[0]
        ax[i].plot(cost, ymax, "ro")
        ax[i].annotate(text=f"$\kappa={int(cost)}$\n{ylabel}={ymax:.1f}",
                       xy=(cost, ymax),
                       xytext=(0.7, 0.1),
                       textcoords="axes fraction",
                       arrowprops=dict(arrowstyle='<|-',
                                       color='red',
                                       lw=1.5,
                                       ls='-'))
        ax[i].set_xlabel("Retraining Cost $\kappa$")
        ax[i].set_yscale("log")

    fig.suptitle(f"Optimal Paramters of different algorithm variants")
    if save_fig:
        plt.savefig(figure_dir /
                    f"{dataset}_{query_dist}_optimized_parameters.pdf",
                    bbox_inches="tight")
    else:
        plt.show()


# %%
def plot_cumulative_strategy_cost(dataset,
                                  query_dist,
                                  request_cost,
                                  save_fig: bool = False,
                                  figure_dir: Path = figures_dir):
    df = load_df(dataset, query_dist, result_base_dir=results_base_dir)
    results_dir = results_base_dir / dataset / query_dist
    matrices = load_cost_matrices(results_dir / "costs.npz")
    offline_C, online_C = matrices["offline_C"],matrices["online_C"]
    retrain_costs = df.retrain_cost.unique()
    cost = get_closest_cost(request_cost, retrain_costs)

    filter_df = df[df.retrain_cost == cost]
    batches = range(len(online_C))
    _C = add_retraining_cost(online_C, cost)
    # index the algorithm
    _df = filter_df.set_index("algorithm")
    # compute the cum cost based on retrains
    _df["cum_strategy_cost"]=_df.retrains_online.apply(
    lambda x: get_cum_costs(x, _C))
    # Also find the cum costs where the retrains occurred 
    _df["retrain_cum_costs"] = _df.apply(lambda row: row["cum_strategy_cost"][row["retrains_online"]],axis=1)

    # first explode the cumulative costs and plot with batches
    _exploded_df = _df.reset_index().groupby("algorithm").cum_strategy_cost.apply(
        lambda x: x.explode().reset_index(name="cum_costs")).reset_index()
    # exploded_df
    fig,ax = plt.subplots()

    sns.lineplot(data=_exploded_df,
                 x="level_1",
                 y="cum_costs",
                 hue="algorithm",
                 style="algorithm",
                 ax=ax,
                 hue_order=hue_order)
    # ax.set_yscale("symlog")
    
    # # now explode the locations where the retrains occurred and their costs
    # _exploded_df = _df[["retrain_cum_costs","retrains_online"]].apply(pd.Series.explode).reset_index()
    # # remove the retrain at t=0
    # _exploded_df = _exploded_df.drop(_exploded_df[_exploded_df.retrains_online==0].index)
    # # plot a scatter plot for retrains
    # sns.scatterplot(data=_exploded_df,
    #             x="retrains_online",
    #             y="retrain_cum_costs",
    #             hue="algorithm",
    #             style="algorithm",
    #             hue_order=hue_order,
    #             s=50,
    #             ax=ax,
    #             legend=False
    #             )
    plt.xlabel("Batches")
    plt.ylabel("Cumulative Strategy Cost")
    plt.title(f"Online Evaluation at $\kappa$={int(cost):g}")
    if save_fig:
        plt.savefig(figure_dir /
                    f"{dataset}_{query_dist}_cum_costs_{int(cost)}.pdf",
                    bbox_inches="tight")
    else:
        plt.show()


def plot_single(df,figure_dir=figures_dir,save_fig=False):
    plot_strat_cost(df,figure_dir,save_fig=save_fig)
    plot_retrains(df,figure_dir,save_fig=save_fig)
    plot_acc(df,figure_dir,save_fig=save_fig)
    plot_acc(df,figure_dir,save_fig=save_fig,query=True)
    
def plot_retrains(df,figure_dir=figures_dir,ax=None,save_fig=False):
    if ax is None:
        fig,ax = plt.subplots(figsize=(5,5))

    plot_column(df,"num_retrains",legend=False,ax=ax)
    ax.set_ylabel("Number of Retrains")
    if save_fig:
        plt.savefig(figure_dir/f"{dataset}_{query_dist}_online_retrains.pdf",bbox_inches="tight")
        plt.close()
    else:
        plt.show()

def plot_strat_cost(df,figure_dir=figures_dir,ax=None,save_fig=False):
    if ax is None:
        fig,ax = plt.subplots(figsize=(5,5))

    plot_column(df,legend=False,ax=ax)
    ylim = get_strategy_cost_ylim(df,online=True)
    ax.set_ylim(bottom=-0.5, top=ylim*1.05)
    ax.set_ylabel("Strategy Cost")
    if save_fig:
        plt.savefig(figure_dir/f"{dataset}_{query_dist}_online_strat_cost.pdf",bbox_inches="tight")
        plt.close()
    else:
        plt.show()

def plot_acc(df,figure_dir=figures_dir,ax=None,save_fig=False,query=False):
    if ax is None:
        fig,ax = plt.subplots(figsize=(5,5))
    if query:
        column = "prequential_query_acc"
        label = "Query Accuracy"
        fname = "query_acc"
    else:
        column = "prequential_acc"
        label = "Data Accuracy"
        fname = "data_acc"
    plot_column(df,column=column,legend=False,ax=ax)
    # ax.set_ylim(0,1)
    ax.set_ylabel(label)
    if save_fig:
        plt.savefig(figure_dir/f"{dataset}_{query_dist}_online_{fname}.pdf",bbox_inches="tight")
        plt.close()
    else:
        plt.show()
# %%

def setup_matplotlib():
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 10,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    "legend.fontsize":10,
    "figure.titlesize": 10,
    })
#%%
def plot_cost_matrices(request_cost=50,save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 10,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":6,
    "figure.titlesize": 10,
    })
    figsize = np.array(set_size(width=fullwidth,subplots=(1,4)))
    figsize[1]*=1.8
    fig = plt.figure(constrained_layout=True,figsize = figsize)
    subfigs = fig.subfigures(1,2)
    dataset_query_dists = [
    ("covcon","data"), # 0
    ("covcon","static_middle"), #1
    ]
    dataset_names = [r"\textsc{CovCon-D}",r"\textsc{CovCon-S}"]
    for (dataset,query_dist),dataset_name,subfig in zip(dataset_query_dists,dataset_names,subfigs):
        ax1,ax2,ax3 = subfig.subplots(1,3,width_ratios=[1,1,0.05])
        df = load_df(dataset,query_dist,verbose=False)
        results_dir = results_base_dir / dataset / query_dist
        matrices = load_cost_matrices(results_dir / "costs.npz")
        offline_C, online_C = matrices["offline_C"],matrices["online_C"]
        offline_iu = np.triu_indices(len(offline_C))
        online_iu = np.triu_indices(len(online_C))
        vmin, vmax = min(offline_C[offline_iu].min(),
                        online_C[online_iu].min()), max(
                            offline_C[offline_iu].max(),
                            online_C[online_iu].max())
        kwargs = {"vmin": vmin, "vmax": vmax}
        retrain_costs = df.retrain_cost.unique()
        cost = get_closest_cost(request_cost, retrain_costs)
        print(f"{cost=}")

        im1 = ax1.imshow(add_retraining_cost(offline_C,cost),**kwargs,cmap="binary")
        ax1.set_xlabel("Batch evaluated ($t$)")
        ax1.set_ylabel("Batch trained ($t^\prime$)")
        ax1.set_title("Offline")
        im2 = ax2.imshow(add_retraining_cost(online_C,cost),**kwargs,cmap="binary")
        ax2.set_xlabel("Batch evaluated")
        ax2.set_title("Online")
        # ax2.set_ylabel("Batch trained")
        subfig.colorbar(im2, cax=ax3)
        # plot the retraining 
        row = df[(df.algorithm == label_rename["Threshold"])
             & (df.retrain_cost == cost)].iloc[0]
        name = label_rename["Threshold"]
        offline_retrains = row["retrains_offline"]
        online_retrains = row["retrains_online"]
        
        T_offline = offline_C.shape[0]-1
        T_online = online_C.shape[0]-1
        args = {"markersize":3}
        _plot_single_retrain(T_offline,offline_retrains,ax1,"r",**args)
        ax1.legend([name], loc="lower left")
        _plot_single_retrain(T_online,online_retrains,ax2,"r",**args)
        ax2.legend([name], loc="lower left")

        subfig.suptitle(dataset_name)
    # fig.subplots_adjust()
    if save_fig:
        plt.savefig(figures_dir/f"covcon_cost_matrices.pdf",bbox_inches="tight")
    else:
        plt.show()

#%%
def plot_half_cost_matrices(request_cost=50,save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':9,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    "figure.titlesize": 7,
    })
    figsize = np.array(set_size(width=columnwidth,subplots=(1,2)))
    figsize[1]*=1.8
    fig,(ax1,ax2) = plt.subplots(
        1,2,
        figsize = figsize,
        sharey=True
        # width_ratios=[1,0.05,1,0.05],
        ) 
    dataset_query_dists = [
    ("covcon","data"), # 0
    ("covcon","static_middle"), #1
    ]
    dataset_names = [r"\textsc{CovCon-D}",r"\textsc{CovCon-S}"]
    for i,((dataset,query_dist),dataset_name,ax) in enumerate(zip(dataset_query_dists,dataset_names,[ax1,ax2])):
        df = load_df(dataset,query_dist,verbose=False)
        results_dir = results_base_dir / dataset / query_dist
        matrices = load_cost_matrices(results_dir / "costs.npz")
        offline_C, online_C = matrices["offline_C"],matrices["online_C"]
        offline_iu = np.triu_indices(len(offline_C))
        online_iu = np.triu_indices(len(online_C))
        vmin, vmax = min(offline_C[offline_iu].min(),
                        online_C[online_iu].min()), max(
                            offline_C[offline_iu].max(),
                            online_C[online_iu].max())
        kwargs = {"vmin": vmin, "vmax": vmax}
        retrain_costs = df.retrain_cost.unique()
        cost = get_closest_cost(request_cost, retrain_costs)
        print(f"{cost=}")

        im1 = ax.imshow(add_retraining_cost(offline_C,cost),**kwargs,cmap="binary")
        ax.set_xlabel("Batch evaluated ($t$)")
        if i==0:
            ax.set_ylabel("Batch trained ($t^\prime$)")
        ax.set_title(dataset_name)
        # im2 = ax2.imshow(add_retraining_cost(online_C,cost),**kwargs,cmap="binary")
        # ax2.set_xlabel("Batch evaluated")
        # ax2.set_title("Online")
        # ax2.set_ylabel("Batch trained")
        fig.colorbar(im1, ax=ax,shrink=0.6)
        # plot the retraining 
        row = df[(df.algorithm == label_rename["Threshold"])
             & (df.retrain_cost == cost)].iloc[0]
        name = label_rename["Threshold"]
        offline_retrains = row["retrains_offline"]
        # online_retrains = row["retrains_online"]
        
        T_offline = offline_C.shape[0]-1
        # T_online = online_C.shape[0]-1
        args = {"markersize":3}
        _plot_single_retrain(T_offline,offline_retrains,ax,"r",**args)
        ax.legend([name], loc="lower left",borderaxespad=0,frameon=False)
        # _plot_single_retrain(T_online,online_retrains,ax2,"r",**args)
        # ax2.legend([name], loc="lower left")

    fig.tight_layout()
    if save_fig:
        plt.savefig(figures_dir/f"covcon_half_cost_matrices.pdf",bbox_inches="tight")
    else:
        plt.show()

#%%
def plot_parameters(save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    dataset = "covcon"
    query_dist = "data"
    df = load_df(dataset, query_dist, result_base_dir=results_base_dir)
    exploded_df = df.join(pd.DataFrame(df["parameters_offline"].to_list()),
                          rsuffix="_param")
    figsize = np.array(set_size(columnwidth,subplots=(1,3)))
    figsize[1]*=2
    fig, ax = plt.subplots(1, 3, figsize=figsize)
    items = ["threshold", "cum_threshold", "period"]
    item_labels = [
        "optimal thresholds", "optimal cumulative thresholds",
        "optimal periodicities"
    ]
    ylabels = [r"$\tau^*$", r"$\tau_{cum}^*$", r"$\phi^*$"]
    for i, (item, label, ylabel) in enumerate(zip(items, item_labels,
                                                  ylabels)):
        sns.lineplot(data=exploded_df, x="retrain_cost", y=item, ax=ax[i])
        ax[i].set_title(ylabel)
        ax[i].set_ylabel("")
        ax[i].set_xlabel("$\kappa$")
        ax[i].set_yscale("log")

    # fig.suptitle(f"Optimal Paramters of different algorithm variants")
    # fig.subplots_adjust()
    fig.subplots_adjust(
        # left=0,
        # right=1,
        # top=1,
        # bottom=0,
        wspace=0.45
        )
    if save_fig:
        plt.savefig(figures_dir /
                    f"{dataset}_{query_dist}_optimized_parameters.pdf",
                    bbox_inches="tight")
    else:
        plt.show()
#%%
def plot_full_online_results(dataset="covcon",query_dist="data",save_fig=False,legend=False,wspace=0.2):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':8,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })
    figsize = np.array(set_size(fullwidth,subplots=(1,3)))
    figsize[1] *= 1.4
    fig,(ax1,ax2,ax3) = plt.subplots(1,3,figsize=figsize)
    df = load_df(dataset,query_dist,verbose=False)
    df = df[df.retrain_cost > 0]
    # get the ylim without baselines
    ylim = get_strategy_cost_ylim(df,online=True)
    sns.lineplot(
        data=df,
        x="retrain_cost",
        y="strategy_cost_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax1,
        palette=default_palette,
        legend=True,
        errorbar="se",
    )
    # ax1.set_yscale("log")
    # ax1.set_xscale("log")
    ax1.set_ylim(top=ylim*1.05)
    sns.lineplot(
        data=df,
        x="retrain_cost",
        y="num_retrains_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax2,
        palette=default_palette,
        legend=False,
        errorbar="ci",
    )
    sns.lineplot(
        data=df,
        x="retrain_cost",
        y="prequential_query_acc_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax3,
        palette=default_palette,
        legend=False,
        errorbar="ci",
    )
    ax1.set_ylabel("")
    ax2.set_ylabel("")
    ax3.set_ylabel("")
    ax1.set_xlabel("$\kappa$")
    ax2.set_xlabel("$\kappa$")
    ax3.set_xlabel("$\kappa$")
    ax1.set_title("Strategy Cost")
    ax2.set_title(r"\# $\texttt{Retrain}$ Decisions")
    ax3.set_title("Query Accuracy")
    
    if legend:
        legend = ax1.legend(loc="upper center",
            bbox_to_anchor=(0.5, -0.2),
            bbox_transform=fig.transFigure,
            ncol=8,
            frameon=False)
        _fig  = legend.figure
        _fig.canvas.draw()
        bbox  = legend.get_window_extent().transformed(_fig.dpi_scale_trans.inverted())
        if save_fig:
            _fig.savefig(figures_dir/"legend.pdf", bbox_inches=bbox)
    
    ax1.get_legend().remove()
    fig.subplots_adjust(wspace=wspace)
    # fig.tight_layout()
    if save_fig:
        plt.savefig(figures_dir / f"{dataset}_{query_dist}_online.pdf",
                    bbox_inches="tight")
    else:
        plt.show()

#%%


def get_best_cara(df):
    _df = df[df.algorithm.isin([label_rename[n] for n in ["Threshold","Periodic","Cumulative Threshold"]])]
    avg_strat_cost = _df.groupby("algorithm").strategy_cost_online.mean()
    best_cara = avg_strat_cost[avg_strat_cost == avg_strat_cost.min()].index[0]
    return best_cara

def plot_synthetic_results(filter_set=["ADWIN","DDM","Oracle","Never Retrain"],
                            save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':8,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })
    filter_set = list(map(label_rename.get,filter_set))
    dataset_query_dists = [
    ("circles","data"), #2
    ("circles","static_middle"), #3
    ("gauss","data"), # 4
    ("gauss","static_middle")] #5
    dataset_names = [
        r"\textsc{Circle-D}",
        r"\textsc{Circle-S}",
        r"\textsc{Gauss-D}",
        r"\textsc{Gauss-S}",]
    figsize = np.array(set_size(fullwidth,subplots=(3,4)))
    figsize[1]*=1.5
    fig = plt.figure(figsize=figsize)
    subfigs = fig.subfigures(1,4)

    for i,((dataset,query_dist),dataset_name,subfig) in enumerate(zip(dataset_query_dists,dataset_names,subfigs)):
        print(dataset,query_dist)
        df = load_df(dataset,query_dist,verbose=False)
        best_cara = get_best_cara(df)
        _filter_set = filter_set + [best_cara]
        print(f"{best_cara=} {_filter_set=}")
        df = df[df.algorithm.isin(_filter_set)]
        # rename variant to Cara
        df.loc[df.algorithm==best_cara,"algorithm"] = r"\textsc{Cara}"
        _hue_order = hue_order.copy()
        _hue_order.append(r"\textsc{Cara}")
        ax1,ax2,ax3 = subfig.subplots(3,1,sharex=True)
        ylim = get_strategy_cost_ylim(df,online=True)
        sns.lineplot(
            data=df,
            x="retrain_cost",
            y="strategy_cost_online",
            hue="algorithm",
            style="algorithm",
            hue_order=_hue_order,
            style_order=_hue_order,
            ax=ax1,
            palette=default_palette,
            legend=False
        )
        ax1.set_ylim(top=ylim*1.1,bottom=-ylim*.1)
        sns.lineplot(
            data=df,
            x="retrain_cost",
            y="num_retrains_online",
            hue="algorithm",
            style="algorithm",
            hue_order=_hue_order,
            style_order=_hue_order,
            ax=ax2,
            palette=default_palette,
            legend=False
        )
        sns.lineplot(
            data=df,
            x="retrain_cost",
            y="prequential_query_acc_online",
            hue="algorithm",
            style="algorithm",
            hue_order=_hue_order,
            style_order=_hue_order,
            ax=ax3,
            palette=default_palette,
            legend=False
        )
        ax3.set_xlabel("$\kappa$")
        ax1.set_ylabel("")
        ax2.set_ylabel("")
        ax3.set_ylabel("")
        ax1.set_title(dataset_name)
        subfig.subplots_adjust(hspace=0.1)
        subfig.align_ylabels([ax1,ax2,ax3])

        if i ==0:
            ax1.set_ylabel("Strategy Cost")
            ax2.set_ylabel(r"\# \texttt{Retrains} Decisions")
            ax3.set_ylabel("Query Accuracy")
        
    if save_fig:
        plt.savefig(figures_dir/"synthetic_results.pdf",bbox_inches="tight")
    else:


        plt.show()
#%%
def plot_real_world_results(filter_set=["ADWIN","DDM","Oracle","Never Retrain"],
                            save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':8,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })
    filter_set = list(map(label_rename.get,filter_set))
    dataset_query_dists = [
    ("elec","data"), #6
    ("airlines","data"), #7
    ("covertype","data"), ]#8 
    dataset_names = [
        r"\textsc{Electricity}",
        r"\textsc{Airlines}",
        r"\textsc{Covertype}"]
    figsize = np.array(set_size(fullwidth,subplots=(3,3)))
    figsize[1]*=1
    fig = plt.figure(figsize=figsize)
    subfigs = fig.subfigures(1,3)

    for i,((dataset,query_dist),dataset_name,subfig) in enumerate(zip(dataset_query_dists,dataset_names,subfigs)):
        print(dataset,query_dist)
        df = load_df(dataset,query_dist,verbose=False)
        best_cara = get_best_cara(df)
        _filter_set = filter_set + [best_cara]
        print(f"{best_cara=} {_filter_set=}")
        df = df[df.algorithm.isin(_filter_set)]
        # rename variant to Cara
        df.loc[df.algorithm==best_cara,"algorithm"] = r"\textsc{Cara}"
        _hue_order = hue_order.copy()
        _hue_order.append(r"\textsc{Cara}")
        ax1,ax2,ax3 = subfig.subplots(3,1,sharex=True)
        ylim = get_strategy_cost_ylim(df,online=True)
        sns.lineplot(
            data=df,
            x="retrain_cost",
            y="strategy_cost_online",
            hue="algorithm",
            style="algorithm",
            hue_order=_hue_order,
            style_order=_hue_order,
            ax=ax1,
            palette=default_palette,
            legend=False
        )
        ax1.set_ylim(top=ylim*1.1,bottom=-ylim*.1)
        sns.lineplot(
            data=df,
            x="retrain_cost",
            y="num_retrains_online",
            hue="algorithm",
            style="algorithm",
            hue_order=_hue_order,
            style_order=_hue_order,
            ax=ax2,
            palette=default_palette,
            legend=False
        )
        sns.lineplot(
            data=df,
            x="retrain_cost",
            y="prequential_query_acc_online",
            hue="algorithm",
            style="algorithm",
            hue_order=_hue_order,
            style_order=_hue_order,
            ax=ax3,
            palette=default_palette,
            legend=False
        )
        ax3.set_xlabel("$\kappa$")
        ax1.set_ylabel("")
        ax2.set_ylabel("")
        ax3.set_ylabel("")
        ax1.set_title(dataset_name)
        subfig.subplots_adjust(hspace=0.1)
        subfig.align_ylabels([ax1,ax2,ax3])
        if i ==0:
            ax1.set_ylabel("Strategy Cost")
            ax2.set_ylabel(r"\# \texttt{Retrains} Decisions")
            ax3.set_ylabel("Query Accuracy")
        
    if save_fig:
        plt.savefig(figures_dir/"real_world_results.pdf",bbox_inches="tight")
    else:
        plt.show()
#%%

def plot_covcon_online_results(dataset="covcon",save_fig=False,legend=True):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':8,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })
    figsize = np.array(set_size(fullwidth,subplots=(2,3)))
    figsize[1] *= 1.2
    fig = plt.figure(constrained_layout=True,figsize=figsize)
    subfigs = fig.subfigures(2,1)
    (ax1,ax2,ax3) = subfigs[0].subplots(1,3)
    (ax4,ax5,ax6)  = subfigs[1].subplots(1,3)
    
    df1 = load_df(dataset,"data",verbose=False)

    # get the ylim without baselines
    ylim = get_strategy_cost_ylim(df1,online=True)
    sns.lineplot(
        data=df1,
        x="retrain_cost",
        y="strategy_cost_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax1,
        palette=default_palette,
        legend=True
    )
    ax1.set_ylim(top=ylim*1.05)
    sns.lineplot(
        data=df1,
        x="retrain_cost",
        y="num_retrains_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax2,
        palette=default_palette,
        legend=False
    )
    sns.lineplot(
        data=df1,
        x="retrain_cost",
        y="prequential_query_acc_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax3,
        palette=default_palette,
        legend=False
    )
    ax1.set_ylabel("")
    ax2.set_ylabel("")
    ax3.set_ylabel("")
    ax1.set_xlabel("")
    ax2.set_xlabel("")
    ax3.set_xlabel("")
    ax1.set_title("Strategy Cost")
    ax2.set_title(r"\# $\texttt{Retrain}$ Decisions")
    ax3.set_title("Query Accuracy")
    subfigs[0].suptitle(r"\textsc{CovCon-D}")
    
    df2 = load_df(dataset,"static_middle",verbose=False)

    # get the ylim without baselines
    ylim = get_strategy_cost_ylim(df2,online=True)
    sns.lineplot(
        data=df2,
        x="retrain_cost",
        y="strategy_cost_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax4,
        palette=default_palette,
        legend=False
    )
    ax4.set_ylim(top=ylim*1.05)
    sns.lineplot(
        data=df2,
        x="retrain_cost",
        y="num_retrains_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax5,
        palette=default_palette,
        legend=False
    )
    sns.lineplot(
        data=df2,
        x="retrain_cost",
        y="prequential_query_acc_online",
        hue="algorithm",
        style="algorithm",
        hue_order=hue_order,
        style_order=hue_order,
        ax=ax6,
        palette=default_palette,
        legend=False
    )
    ax4.set_ylabel("")
    ax5.set_ylabel("")
    ax6.set_ylabel("")
    ax4.set_xlabel("$\kappa$")
    ax5.set_xlabel("$\kappa$")
    ax6.set_xlabel("$\kappa$")
    # ax4.set_title("Strategy Cost")
    # ax5.set_title(r"\# $\texttt{Retrain}$ Decisions")
    # ax6.set_title("Query Accuracy")
    subfigs[1].suptitle(r"\textsc{CovCon-S}")
    
    
    if legend:
        legend = ax1.legend(loc="upper center",
            bbox_to_anchor=(0.5, -0.2),
            bbox_transform=fig.transFigure,
            ncol=4,
            frameon=False)
        _fig  = legend.figure
        _fig.canvas.draw()
        bbox  = legend.get_window_extent().transformed(_fig.dpi_scale_trans.inverted())
        if save_fig:
            _fig.savefig(figures_dir/"legend.pdf", bbox_inches=bbox)
    ax1.get_legend().remove()
    # fig.tight_layout()
    if save_fig:
        plt.savefig(figures_dir / f"{dataset}_online.pdf",
                    bbox_inches="tight")
    else:
        plt.show()
#%%

def pivot_results(df,metric):
    _df = df.loc[:,["algorithm","model_seed",metric,"retrain_cost"]]
    _df = _df.pivot(["retrain_cost","model_seed"],columns=["algorithm"],values=[metric])
    # drop mltilevel columsn
    _df.columns = _df.columns.droplevel()
    # drop retrain cost = 0
    _df = _df.drop(index=0)
    return _df

def compute_mpe(dataset,query_dist,metric="strategy_cost_online"):
    df = load_df(dataset,query_dist,label_mapping=None)
    _df = pivot_results(df,metric)
    oracle = _df.pop("Oracle")
    _df = _df.subtract(oracle,axis="index")
    _df = _df.divide(oracle,axis="index")
    # make absolute to have Mean Absolute Percentage Error 
    _df = _df.apply(np.abs)
    _df = _df*100 # Percentage Error
    # Mean Percentage error wrt oracle 
    _df = _df.groupby(level=0).mean()

    mpe = _df.mean()
    ci_95 = _df.sem()*1.96
    # sns.barplot(data=_df[['Cumulative Threshold', 'Markov', 'Never Retrain',
    #    'Periodic', 'Threshold']])
    # plt.ylabel("Strategy MPE")
    # plt.xticks(rotation=45)
    return mpe,ci_95

def create_mpe_table(version="synthetic", mask = True, save=False):
    fname = ""
    if version == "synthetic":
        _dataset_query_dists = dataset_query_dists[:6]
        _dataset_names = dataset_names[:6]
        fname +="synthetic"
    elif version == "real-world":
        _dataset_query_dists = dataset_query_dists[6:]
        _dataset_names = dataset_names[6:]
        fname +="real_world"
    elif version == "full":
        fname+="full"
        _dataset_query_dists= dataset_query_dists
        _dataset_names = dataset_names
    dfs = []

    for (dataset,query),name in zip(_dataset_query_dists,_dataset_names):
        mpe,ci_95 = compute_mpe(dataset,query)
        mpe["Dataset"] = name
        dfs.append(mpe)

    df = pd.concat(dfs,axis=1)
    df = df.T.set_index("Dataset")
    if mask:
        fname+="_small"
        df = df[["Threshold","Cumulative Threshold","Periodic"]]
    else:
        df = df[["Threshold","Cumulative Threshold","Periodic","Never Retrain","Markov","ADWIN","DDM"]]

    df.columns = df.columns.map(label_rename.get)
    df = df.convert_dtypes()
    df.columns.name="Dataset"
    df.index.name=""
    df = df.round(2)
    fname += ".tex"
    print(f"Saving table to {tables_dir/fname}")
    if save:
        df.to_latex(tables_dir/fname,escape=False)
    else:
        print(df.to_latex(escape=False))

def compute_query_acc(dataset,query):
    _df = load_df(dataset,query,label_mapping=None)
    _df = pivot_results(_df,"prequential_query_acc_online")
    _df = _df.groupby(level=0).mean()
    acc = _df.mean()
    ci_95 = _df.sem()*1.96
    return acc,ci_95

def create_acc_table(version="synthetic", mask = True, save=False):
    fname = ""
    if version == "synthetic":
        _dataset_query_dists = dataset_query_dists[:6]
        _dataset_names = dataset_names[:6]
        fname +="synthetic"
    elif version == "real-world":
        _dataset_query_dists = dataset_query_dists[6:]
        _dataset_names = dataset_names[6:]
        fname +="real_world"
    elif version == "full":
        fname+="full"
        _dataset_query_dists= dataset_query_dists
        _dataset_names = dataset_names
    dfs = []

    for (dataset,query),name in zip(_dataset_query_dists,_dataset_names):
        acc,ci_95 = compute_query_acc(dataset,query)
        best_cara = acc[["Threshold","Cumulative Threshold","Periodic"]].idxmax()
        acc["Dataset"] = name
        acc["Cara"] = acc[best_cara]
        dfs.append(acc)

    df = pd.concat(dfs,axis=1)
    df = df.T.set_index("Dataset")
    if mask:
        fname+="_small"
        df = df[["Threshold","Oracle","ADWIN"]]
    else:
        df = df[["Threshold","Cumulative Threshold","Periodic","Oracle","Never Retrain","Markov","ADWIN","DDM"]]

    df.columns = df.columns.map(label_rename.get)
    df = df.convert_dtypes()
    df.columns.name="Dataset"
    df.index.name=""
    df = df.round(2)
    fname += "_query_acc.tex"
    print(f"Saving table to {tables_dir/fname}")
    if save:
        df.to_latex(tables_dir/fname,escape=False)
        df.to_markdown((tables_dir/fname).with_suffix('.md'))
    else:
        print(df.to_latex(escape=False))
    
def compute_num_retrains(dataset,query):
    _df = load_df(dataset,query,label_mapping=None)
    _df = pivot_results(_df,"num_retrains_online")
    _df = _df.groupby(level=0).mean()
    num_retrains = _df.mean()
    ci_95 = _df.sem()*1.96
    return num_retrains,ci_95

def create_retrains_table(version="synthetic", mask = True, save=False):
    fname = ""
    if version == "synthetic":
        _dataset_query_dists = dataset_query_dists[:6]
        _dataset_names = dataset_names[:6]
        fname +="synthetic"
    elif version == "real-world":
        _dataset_query_dists = dataset_query_dists[6:]
        _dataset_names = dataset_names[6:]
        fname +="real_world"
    elif version == "full":
        fname+="full"
        _dataset_query_dists= dataset_query_dists
        _dataset_names = dataset_names
    dfs = []

    for (dataset,query),name in zip(_dataset_query_dists,_dataset_names):
        num_retrains,ci_95 = compute_num_retrains(dataset,query)
        num_retrains["Dataset"] = name
        dfs.append(num_retrains)

    df = pd.concat(dfs,axis=1)
    df = df.T.set_index("Dataset")
    if mask:
        fname+="_small"
        df = df[["Threshold","Oracle","ADWIN"]]
    else:
        df = df[["Threshold","Cumulative Threshold","Periodic","Oracle","Never Retrain","Markov","ADWIN","DDM"]]

    df.columns = df.columns.map(label_rename.get)
    df = df.convert_dtypes()
    df.columns.name="Dataset"
    df.index.name=""
    df = df.round(2)
    fname += "_num_retrains.tex"
    print(f"Saving table to {tables_dir/fname}")
    if save:
        df.to_latex(tables_dir/fname,escape=False)
    else:
        print(df.to_latex(escape=False))
    

def compute_num_retrains(dataset,query):
    _df = load_df(dataset,query,label_mapping=None)
    _df = pivot_results(_df,"num_retrains_online")
    _df = _df.groupby(level=0).mean()
    num_retrains = _df.mean()
    ci_95 = _df.sem()*1.96
    return num_retrains,ci_95

def create_retrains_table(version="synthetic", mask = True, save=False):
    fname = ""
    if version == "synthetic":
        _dataset_query_dists = dataset_query_dists[:6]
        _dataset_names = dataset_names[:6]
        fname +="synthetic"
    elif version == "real-world":
        _dataset_query_dists = dataset_query_dists[6:]
        _dataset_names = dataset_names[6:]
        fname +="real_world"
    elif version == "full":
        fname+="full"
        _dataset_query_dists= dataset_query_dists
        _dataset_names = dataset_names
    dfs = []

    for (dataset,query),name in zip(_dataset_query_dists,_dataset_names):
        num_retrains,ci_95 = compute_num_retrains(dataset,query)
        num_retrains["Dataset"] = name
        dfs.append(num_retrains)

    df = pd.concat(dfs,axis=1)
    df = df.T.set_index("Dataset")
    if mask:
        fname+="_small"
        df = df[["Threshold","Oracle","ADWIN"]]
    else:
        df = df[["Threshold","Cumulative Threshold","Periodic","Oracle","Never Retrain","Markov","ADWIN","DDM"]]

    df.columns = df.columns.map(label_rename.get)
    df = df.convert_dtypes()
    df.columns.name="Dataset"
    df.index.name=""
    df = df.round(2)
    fname += "_num_retrains.tex"
    print(f"Saving table to {tables_dir/fname}")
    if save:
        df.to_latex(tables_dir/fname,escape=False)
    else:
        print(df.to_latex(escape=False))


def _plot_trade_off(dataset,query_dist,ax=None,label_rename=label_rename):
    df = load_df(dataset,query_dist,label_mapping=None)
    df = df.groupby(["algorithm","model_training_time_online"]).prequential_query_acc_online.mean().reset_index()
    df = df[df.model_training_time_online>0]
    # figsize = np.array(set_size(columnwidth,subplots=(1.8,1)))
    # fig, (ax, leg_ax) = plt.subplots(2, 1, height_ratios=[1, 0.5],figsize=figsize)
    if ax is None:
        fig,ax =plt.subplots()
    for algorithm in [
        "Threshold",
        # "Cumulative Threshold",
        # "Periodic",
        # "Threshold",
        # "Oracle",
        # "Markov"
    ]:
        _df = df[df.algorithm == algorithm].sort_values("model_training_time_online")
        _df = _df.replace(label_rename)
        algorithm_name = label_rename[algorithm]
        color = default_palette[algorithm_name]
        ax.plot(_df.model_training_time_online,_df.prequential_query_acc_online,label=algorithm_name,color=color,linewidth=2)
    
    for baseline_algorithm in [
        "ADWIN",
        "Incremental",
        "ADWIN+Incremental",
        "Adaptive Incremental",
        "Always Retrain",
        # "Never Retrain"
    ]:
        _df = df[df.algorithm == baseline_algorithm].sort_values("model_training_time_online")
        _df = _df.replace(label_rename)
        algorithm_name = label_rename[baseline_algorithm]
        color = default_palette[algorithm_name]
        ax.scatter(_df.model_training_time_online,_df.prequential_query_acc_online,label=algorithm_name,marker='s',color=color)
    
    ax.set_ylabel("Average Query Accuracy")
    ax.set_xlabel("Total Time Retraining/Updating model (in seconds)")
    return ax



def plot_trade_off(dataset,query_dist,save_fig:bool = False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })
    df = load_df(dataset,query_dist,label_mapping=None)
    df = df.groupby(["algorithm","model_training_time_online"]).prequential_query_acc_online.mean().reset_index()
    df = df[df.model_training_time_online>0]
    figsize = np.array(set_size(columnwidth,subplots=(1.8,1)))
    fig, (ax, leg_ax) = plt.subplots(2, 1, height_ratios=[1, 0.5],figsize=figsize)
    for algorithm in [
        "Threshold",
        # "Cumulative Threshold",
        # "Periodic",
        # "Threshold",
        # "Oracle",
        # "Markov"
    ]:
        _df = df[df.algorithm == algorithm].sort_values("model_training_time_online")
        _df = _df.replace(label_rename)
        algorithm_name = label_rename[algorithm]
        color = default_palette[algorithm_name]
        ax.plot(_df.model_training_time_online,_df.prequential_query_acc_online,label=algorithm_name,color=color)
    
    for baseline_algorithm in [
        "ADWIN",
        "Incremental",
        "ADWIN+Incremental",
        "Adaptive Incremental",
        "Always Retrain"
    ]:
        _df = df[df.algorithm == baseline_algorithm].sort_values("model_training_time_online")
        _df = _df.replace(label_rename)
        algorithm_name = label_rename[baseline_algorithm]
        color = default_palette[algorithm_name]
        ax.scatter(_df.model_training_time_online,_df.prequential_query_acc_online,label=algorithm_name,marker='s',color=color)
    
    ax.set_ylabel("Average Query Accuracy")
    ax.set_xlabel("Total Time Retraining/Updating model (in seconds)")
    ax.legend()
    legend = ax.get_legend_handles_labels()
    ax.get_legend().remove()
    leg_ax.axis("off")
    leg_ax.legend(*legend, ncol=2,loc="lower center",borderaxespad=0)
    ax.set_xscale("log")
    ax.set_title(DATASET_MAPPING[(dataset,query_dist)])
    fig.subplots_adjust(right=1,hspace=0.05)


def plot_row_trade_offs(dataset_query_ids = [0,2,4]):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })      
    figsize = np.array(set_size(fullwidth,subplots=(2,3)))
    # fig,(leg_ax,axes) = plt.subplots(2,3,,figsize=figsize)
    fig = plt.figure(figsize=figsize)
    gs = fig.add_gridspec(2, 3,height_ratios=[0.1,1])
    axes = []
    leg_ax = fig.add_subplot(gs[0, :])
    axes.append(fig.add_subplot(gs[1, 0]))
    axes.append(fig.add_subplot(gs[1, 1]))
    axes.append(fig.add_subplot(gs[1, 2]))

    for i,(ax,(dataset,query)) in enumerate(zip(axes,np.array(dataset_query_dists)[dataset_query_ids])):
        ax = _plot_trade_off(dataset,query,ax=ax)
        ax.set_xscale("log")
        ax.set_xlabel("")
        if i ==0 :
            ax.set_ylabel("Average Query Accuracy")
        else:
            ax.set_ylabel("")
        ax.set_title(DATASET_MAPPING[(dataset,query)])

    ax.legend()
    legend = ax.get_legend_handles_labels()
    ax.get_legend().remove()
    leg_ax.axis("off")
    leg_ax.legend(*legend, ncol=3,loc="lower center",borderaxespad=0,frameon=False,mode="expand")
    fig.supxlabel("Total Time Retraining/Updating model (in seconds)")
    fig.tight_layout()

def plot_grid_trade_offs(save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 8,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":8,
    "ytick.labelsize":8,
    "legend.fontsize":8,
    })      
    figsize = np.array(set_size(fullwidth,subplots=(4,3)))
    # fig,(leg_ax,axes) = plt.subplots(2,3,,figsize=figsize)
    fig = plt.figure(figsize=figsize)
    # gs = fig.add_gridspec(4, 3,height_ratios=[0.01,1,1,1])
    # axes = []
    # leg_ax = fig.add_subplot(gs[0, :])
    # axes.append(fig.add_subplot(gs[1, 0]))
    # axes.append(fig.add_subplot(gs[1, 1]))
    # axes.append(fig.add_subplot(gs[1, 2]))
    # axes.append(fig.add_subplot(gs[2, 0]))
    # axes.append(fig.add_subplot(gs[2, 1]))
    # axes.append(fig.add_subplot(gs[2, 2]))
    # axes.append(fig.add_subplot(gs[3, 0]))
    # axes.append(fig.add_subplot(gs[3, 1]))
    # axes.append(fig.add_subplot(gs[3, 2]))
    leg_fig,subfig = fig.subfigures(2,1,height_ratios=[0.1,1])
    leg_ax = leg_fig.subplots()
    axes = subfig.subplots(3,3,squeeze=True)
    axes = axes.flatten()
    label_rename = OrderedDict({
    "Always Retrain":r"\textsc{Always Retrain}",
    "Incremental": r"\textsc{Always Update}",
    "ADWIN": r"\textsc{ADWIN-Retrain}",
    "ADWIN+Incremental": r"\textsc{ADWIN-Update}",
    "Threshold": r"\textsc{Cara-T-Retrain}",
    "Adaptive Incremental": r"\textsc{HAT-Update}",
    })
    dataset_query_dists = [
    ("covcon","data"), # 0
    ("circles","data"), #2
    ("gauss","data"), # 4
    ("covcon","static_middle"), #1
    ("circles","static_middle"), #3
    ("gauss","static_middle"), #5
    ("elec","data"), #6
    ("airlines","data"), #7
    ("covertype","data"), #8
    ]

    for i,(ax,(dataset,query)) in enumerate(zip(axes,np.array(dataset_query_dists))):
        ax = _plot_trade_off(dataset,query,ax=ax,label_rename=label_rename)
        ax.set_xscale("log")
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_ylim(0.45,1.05)
        # if i % 3 != 0:
            # ax.set_yticklabels([])
        ax.set_title(DATASET_MAPPING[(dataset,query)])

    ax.legend()
    legend_handles,legend_labels = ax.get_legend_handles_labels()
    order = [legend_labels.index(l) for l in label_rename.values()]
    legend_labels = np.array(legend_labels)[order]
    legend_handles = np.array(legend_handles)[order]
    ax.get_legend().remove()
    leg_ax.axis("off")
    leg_ax.legend(legend_handles,legend_labels, ncol=3,loc="lower center",borderaxespad=0,frameon=False,mode="expand")
    subfig.supylabel("Average Query Accuracy",fontsize="x-large")
    subfig.supxlabel("Total Time Retraining/Updating model (in seconds)",fontsize="x-large")
    leg_fig.subplots_adjust(right=0.8,bottom=0.1,left=0.05,top=0.95)
    subfig.subplots_adjust(wspace=0.25,hspace=0.45,right=0.85,bottom=0.1,left=0.1,top=0.95)
    ax.text(
        0.9,0.2,
        "Real-World",
        rotation="vertical",
        transform=subfig.transFigure,
        va="center",ha="right",fontsize="x-large"
    )
    ax.text(
        0.9,0.475,
        "Synthetic",
        rotation="vertical",
        transform=subfig.transFigure,
        va="center",ha="right",fontsize="x-large"
    )
    ax.text(
        0.9,0.775,
        "Synthetic",
        rotation="vertical",
        transform=subfig.transFigure,
        va="center",ha="right",fontsize="x-large"
    )
    if save_fig:
        fig.savefig(figures_dir/"incremental-trade-off-grid.pdf",bbox_inches="tight",pad_inches=0)
#%%
if __name__ == "__main__":


    # dataset_query_dists = [
    #("covcon","data"), # 0
    #("covcon","static_middle"), #1
    #("circles","data"), #2
    #("circles","static_middle"), #3
    #("gauss","data"), # 4
    #("gauss","static_middle"), #5
    #("elec","data"), #6
    #("airlines","data"), #7
    #("covertype","data"), #8
    # ]

    # figures_dir = project_root/"paper"/"ICML_2023"/"figs"
    # results_base_dir = project_root/"results"
    save_fig = False
    #%%
    # plot_cost_matrices(request_cost=20,save_fig=save_fig)
    # plot_half_cost_matrices(request_cost=20,save_fig=save_fig)
    # plot_parameters(save_fig=save_fig)

    results_base_dir = project_root / "results"
    plot_full_online_results("covcon","data",legend=True,save_fig=save_fig)
    #%%
    # plot_full_online_results("covcon","static_middle",save_fig=save_fig)

    # plot_full_online_results("circles","data",save_fig=save_fig,wspace=0.25)
    # plot_full_online_results("circles","static_middle",save_fig=save_fig,wspace=0.22)

    # plot_full_online_results("gauss","data",save_fig=save_fig,wspace=0.22)
    # plot_full_online_results("gauss","static_middle",save_fig=save_fig,wspace=0.22)

    # plot_full_online_results("airlines","data",save_fig=save_fig,wspace=0.22)
    # plot_full_online_results("elec","data",save_fig=save_fig,wspace=0.22)
    # plot_full_online_results("covertype","data",save_fig=save_fig,wspace=0.22)
    # plot_synthetic_results(save_fig)
    # plot_real_world_results(save_fig)

    create_mpe_table("synthetic",mask=True,save=save_fig)
    create_mpe_table("real-world",mask=True,save=save_fig)
    create_mpe_table("full",mask=False,save=save_fig)

    create_acc_table("synthetic",mask=True,save=save_fig)
    create_acc_table("real-world",mask=True,save=save_fig)
    create_acc_table("full",mask=False,save=save_fig)

    create_retrains_table("synthetic",mask=True,save=save_fig)
    create_retrains_table("real-world",mask=True,save=save_fig)
    create_retrains_table("full",mask=False,save=save_fig)

    # %%
    # for dataset,query_dist in dataset_query_dists:
    #     plot_trade_off(dataset,query_dist,save_fig=False)
    results_base_dir = project_root / "results_HT"
    plot_grid_trade_offs(save_fig=save_fig)
# %%
# dataset,query = "elec","data"
# plt.rcParams.update({
#     'text.usetex':True,
#     # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
#     'font.size': 12,
#     'font.family': 'Times New Roman',
#     'axes.titlesize':12,
#     'figure.dpi':100,
#     "xtick.labelsize":12,
#     "ytick.labelsize":12,
#     "legend.fontsize":12,
#     })  
# fig,(leg_ax,ax) = plt.subplots(2,1,height_ratios=[0.2,1],figsize=(5,5))
# _plot_trade_off(dataset,query,ax=ax)
# ax.set_xscale("log")
# ax.legend()
# legend = ax.get_legend_handles_labels()
# ax.get_legend().remove()
# leg_ax.axis("off")
# leg_ax.legend(*legend, ncol=2,loc="lower center",borderaxespad=0,frameon=False)
# ax.set_ylabel("Average Query Accuracy")
# ax.set_xlabel("Total Time Retraining/Updating model (in seconds)")
# ax.set_title(DATASET_MAPPING[(dataset,query)])
# fig.tight_layout()
# fig.savefig(f"/Users/mahadeva/Downloads/{dataset}_{query}_trade_off_plot.pdf",bbox_inches="tight",pad_inches=0)
# %%

# %%
