import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.metrics.pairwise import cosine_similarity, rbf_kernel, euclidean_distances
from typing import Dict, List, Optional, Callable, Tuple, Union
from scipy.interpolate import RBFInterpolator,NearestNDInterpolator, LinearNDInterpolator
from functools import partial
from sklearn.metrics import accuracy_score
Classifier = Union[SGDClassifier, LogisticRegression, RandomForestClassifier]

def query_data_model_overlap(model:Classifier,data:dict,eps:float=1e-15,query_data:dict=None):
    """Return Query Data Model overlap loss

    Args:
        model (Classifier): the trained model
        data (dict): the data dictionary
        eps (float, optional): clipping value for handling -inf. Defaults to 1e-15.
        query_data (dict,optional): An explicit data dictionary from which queries should be used. 
                                    Used when a different query and training data 
                                    are to required. Default None will use queries in `data` 
                                    argument
    """
    X_new = data["X_train"]
    y_new = data["y_train"]
    if query_data is not None:
        X_query = query_data["X_query"]
    # otherwise use the queries from the data dict
    else:
        X_query = data["X_query"]
    y_new_onehot = np.zeros((y_new.size,y_new.max()+1))
    y_new_onehot[np.arange(y_new.size),y_new]=1
    # y_new = np.
    y_pred = model.predict_proba(X_new)
    # clip probabilities close to 1 or 0 
    y_pred = np.clip(y_pred,eps,1-eps)
    # renormalize the probabilities
    y_pred /= y_pred.sum(axis=1)[:, np.newaxis]
    # compute cross entropy loss of current model for each data point
    ce_loss = -(y_new_onehot * np.log(y_pred)).sum(axis=1)
    # compute pairwise similarities between queries and data
    # sim = cosine_similarity(X_query,X_new)
    sim = rbf_kernel(X_query,X_new)

    # find the average query, data, model overlapp loss
    # lower is better
    query_data_model_overlap = (sim@ce_loss/ce_loss.shape[0]).sum()
    
    # # query entropy
    # # gather query probs
    # query_pred = model.predict_proba(X_query)
    # # clip query probs
    # query_pred = np.clip(query_pred,eps,1-eps)
    # # renormalize query probs
    # query_pred /= query_pred.sum(axis=1)[:,np.newaxis]
    # # calculate self entropy
    # query_entropy = -(query_pred*np.log(query_pred)).sum(axis=1)
    
    # return ce_loss.mean()+query_entropy.mean()
    return query_data_model_overlap

def bayes_query_data_model_loss(model:Classifier,data:dict,gamma:Optional[float]=None,query_data:Optional[dict]=None):
    """Return 0/1 Query Data Model overlap loss

    Args:
        model (Classifier): the trained model
        data (dict): the data dictionary
        gamma (float, optional): The hyperparameter for the RBF kernel. Defaults to None.
        query_data (dict,optional): An explicit data dictionary from which queries should be used. 
                                    Used when a different query and training data 
                                    are to required. Default None will use queries in `data` 
                                    argument
    """
    X_new = data["X_train"]
    y_new = data["y_train"]
    if query_data is not None:
        X_query = query_data["X_query"]
    # otherwise use the queries from the data dict
    else:
        X_query = data["X_query"]
    y_pred = model.predict(X_new)
    zero_one_loss = (y_pred!=y_new)
    # compute pairwise similarities between queries and data
    
    # sim = cosine_similarity(X_query,X_new)
    # sim[sim<0]=0
    sim = rbf_kernel(X_query,X_new,gamma=gamma)
    # print(sim.mean(axis=1))
    # find the estimated query misclassifications 
    # lower is better
    query_data_model_overlap = ((sim@zero_one_loss)/zero_one_loss.shape[0]).sum()
    return query_data_model_overlap

def bayes_query_data_model_loss_with_onehot(model:Classifier,data:dict,gamma:Optional[float]=None,query_data:Optional[dict]=None):
    """Return 0/1 Query Data Model overlap loss

    Args:
        model (Classifier): the trained model
        data (dict): the data dictionary. Expects embeddings for computation.
        gamma (float, optional): The hyperparameter for the RBF kernel. Defaults to None.
        query_data (dict,optional): An explicit data dictionary from which queries should be used. 
                                    Used when a different query and training data 
                                    are to required. Default None will use queries in `data` 
                                    argument
    """
    # Use categorical data for prediction
    X_new = data["X_train"]
    y_new = data["y_train"]
    y_pred = model.predict(X_new)
    zero_one_loss = (y_pred!=y_new)

    # Get one hot encoded queries 
    if query_data is not None:
        X_query = query_data["X_onehot_query"]
    # otherwise use the queries from the data dict
    else:
        X_query = data["X_onehot_query"]
    # get one hot encoded data
    X_onehot_data = data["X_onehot_train"]
    # find pairwise similarity with one-hot encoded features
    sim = rbf_kernel(X_query,X_onehot_data,gamma=gamma)
    # find the estimated query misclassifications
    # lower is better
    query_data_model_overlap = ((sim@zero_one_loss)/zero_one_loss.shape[0]).sum()
    return query_data_model_overlap

def discrete_query_data_model_loss(model:Classifier,data:dict,bins:np.ndarray,query_data:dict=None):
    """Return 0/1 discrete Query Data Model overlap loss

    Args:
        model (Classifier): the trained model
        data (dict): the data dictionary
        bins (np.ndarray): The bins for each dimension of data.
        query_data (dict,optional): An explicit data dictionary from which queries should be used. 
                                    Used when a different query and training data 
                                    are to required. Default None will use queries in `data` 
                                    argument
    """
    X_new = data["X_train"]
    y_new = data["y_train"]
    if query_data is not None:
        X_query = query_data["X_query"]
    # otherwise use the queries from the data dict
    else:
        X_query = data["X_query"]
    y_pred = model.predict(X_new)
    misclassifications = (y_pred!=y_new)
    # find the count of misclassifications in each bin
    misclass_hist, _ = np.histogramdd(X_new,bins=bins,weights=misclassifications)
    # find the number of data points in each bin
    hist,_ = np.histogramdd(X_new,bins=bins)
    # average the regions with non zero misclassification
    non_zero_misclass = misclass_hist>0
    misclass_hist[non_zero_misclass]/=hist[non_zero_misclass]


    # find the bins into which the queries land
    query_hist,_ = np.histogramdd(X_query,bins=bins)


    # find the estimated query misclassifications in each bin by multiplying 
    # sum over all bins to get overall estimate  
    query_data_model_overlap = (query_hist*misclass_hist).sum()
    return query_data_model_overlap


def discretize_labels_RBF(data:dict,bins:List[np.ndarray],grid:Optional[np.ndarray]=None):
    """Function that returns the labels values in a discretized feature 
    space using an RBF interpolator. Slower for large scale data.

    Args:
        data (dict): The dictionary that contains data X_train and labels y_train
        bins (List[np.ndarray]): The bins for each dimension of data. Can be generated from np.linspace.
        grid (np.ndarray,optional): The pre-computed numpy meshgrid from the centres of 
        bins. Defaults to None.

    Returns:
        np.ndarray: the discretised labels for X_train data
    """
    D = data["X_train"].shape[1]
    # compute grid if not given
    if grid is None:
        # make the centres of the bins in each dimension
        bin_centres = [(bin[1:]+bin[:-1])/2 for bin in bins]
        # use meshgrid to make into a grid
        grid = np.array(np.meshgrid(*bin_centres))
    
    # convert from 0-1 to -1,1
    y = 2*data["y_train"]-1
    # create the RBF interpolater on the data available  
    interp = RBFInterpolator(data["X_train"],y,kernel="gaussian",epsilon=1)
    # interpolate the grid values 
    grid_values = interp(grid.reshape(D,-1).T)
    return grid_values

def discretize_labels_hist(data:dict,bins:List[np.ndarray]):
    """Function that returns the labels values in a discretized data space
    using histograms and bins for each dimension.

    Args:
        data (dict): the dictionary that contains the data X_train and labels y_train.
        bins (List[np.ndarray]): The bins for each dimension of data. Can be generated from np.linspace

    Returns:
        np.ndarray: the discretized data
    """
    # convert from 0,1 to -1,1
    y = 2*data["y_train"]-1
    # weigh the histogram with the label values to get an estimate for each bin location
    hist, _ = np.histogramdd(data["X_train"],bins=bins,weights=y)
    # return the weighted label histogram as the estimate
    return hist.flatten()

def compute_query_misclassifications(model:Classifier,model_data:dict,new_data:dict,query_data:dict=None):
    if query_data is not None:
        _X_query = query_data["X_query"]
        # if no ground exists return NaN
        if "y_query" not in query_data:
            return np.nan
        else:
            _y_query = query_data["y_query"]
    else:
        _X_query = new_data["X_query"]
        _y_query = new_data["y_query"]
    query_preds = model.predict(_X_query)
    query_incorrect = (query_preds != _y_query).sum()
    return query_incorrect

def compute_data_accuracy(model:Classifier,model_data:dict,new_data:dict,query_data:dict=None):
    _X_train = new_data["X_train"]
    _y_train = new_data["y_train"]
    preds = model.predict(_X_train)
    acc = accuracy_score(_y_train,preds)
    return acc

def compute_query_accuracy(model:Classifier,model_data:dict,new_data:dict,query_data:dict=None):
    if query_data is not None:
        _X_query = query_data["X_query"]
        # if no ground exists return NaN
        if "y_query" not in query_data:
            return np.nan
        else:
            _y_query = query_data["y_query"]
    else:
        _X_query = new_data["X_query"]
        _y_query = new_data["y_query"]
    query_preds = model.predict(_X_query)
    query_acc = accuracy_score(_y_query,query_preds)
    return query_acc


def compute_qdm_diff(model:Classifier,model_data:dict,new_data:dict,query_data:dict,qdm_fn:Optional[Callable]=bayes_query_data_model_loss,**kwargs):
    """Function to compute QDM difference measures.

    Args:
        model (SGDClassifier): The current model instance
        model_data (dict): The data dictionary used to train the current model
        new_data (dict): THe data dictionary corresponding to the incoming data
        query_data (dict): The dictionary corresponding to incoming queries
        qdm_fn (Optional[Callable], optional): The function to compute a single QDM. Defaults to bayes_query_data_model_loss.

    Returns:
        float: The difference between the QDM computed
    """
    # QDM with current model, new data and queries
    QDM_t_t = qdm_fn(
        model=model,
        data=new_data,
        query_data=query_data,
        **kwargs
        )
    # print(f"{QDM_t_t=}")
    # QDM with current model, data used to train current model and queries
    QDM_t_tprime = qdm_fn(
        model=model,
        data=model_data,
        query_data=query_data,
        **kwargs
        )
    # print(f"{QDM_t_tprime=}")

    return QDM_t_t - QDM_t_tprime
    