import argparse
import os
from html import parser
from pathlib import Path

import numpy as np


proj_root = Path(__file__).parent.parent.resolve()
import logging
import shutil
from importlib import import_module
from time import perf_counter as time

import hydra
import pandas as pd
import ray
from hydra.core.hydra_config import HydraConfig
from hydra.utils import get_original_cwd
from omegaconf import DictConfig, OmegaConf

from get_running_times import gather_total_time
from analyse import join_dfs, offline_optimize, online_execute, save_dfs
from baseline import gather_baselines, gather_incremental_baselines
from qdm import compute_data_accuracy, compute_query_accuracy
from utils import (_compute_prequential_accuracy, _compute_strategy_cost,
                   compute_cost_matrix_multiple, get_model_dict, import_obj,
                   save_matrices,load_model_dicts,load_cost_matrices)

log = logging.getLogger(__name__)

@hydra.main(version_base=None,config_path="config")
def run(cfg:DictConfig)->None:
    # turn on Ray for parallel processing
    if cfg.use_ray:
        ray.init(num_cpus=cfg.ray.num_cpus,ignore_reinit_error=True)
    log.info(f"Performing Analysis for {cfg.dataset.name} dataset")
    # load the data
    load_fn = import_obj(cfg.dataset.load_fn)
    offline_datas, online_datas, offline_queries, online_queries = load_fn(cfg)
    log.info("Gathered data and queries")

    results_dir = proj_root / cfg.results_dir
    if "random_state" in cfg.model.kwargs:
        model_seed = cfg.model.kwargs.random_state
        seed_results_dir:Path = results_dir/f"model_seed_{model_seed}"
    else:
        seed_results_dir = results_dir
    # load the pre-computed cost data
    costs_file = seed_results_dir/f"costs.npz"
    if costs_file.exists():
        costs = load_cost_matrices(costs_file)  
    else:
        raise FileNotFoundError(f"Costs file not found at {costs_file}. Please compute the cost matrix prior to analysis.")
    log.info("Loaded Cost Matrices")
    # Load the trained models 
    models_file = seed_results_dir/f"models.joblib"
    if models_file.exists():
        offline_models,online_models = load_model_dicts(models_file,use_ray=cfg.use_ray)
    else:
        raise FileNotFoundError(f"Models file not found at {models_file}. Please run `python create_cost_matrices.py` before analysis.")
    log.info("Loaded Trained Models")

    max_retrain_cost_file = results_dir/"max_retrain_cost.txt"
    if max_retrain_cost_file.exists():
        with max_retrain_cost_file.open("r") as fp:
            max_retrain_cost = int(fp.read())
    else:
        raise FileNotFoundError(f"Max retrain cost file not found at {max_retrain_cost_file}. Please run `python find_max_retraining_cost.py` before analysis.")
    log.info(f"Loaded Max Retrain Cost as {max_retrain_cost}")

    # Load the training time estimates
    model_training_time_file = results_dir/"model_training_time.txt"
    if model_training_time_file.exists():
        with model_training_time_file.open('r') as fp:
            model_training_time = float(fp.read())
    else:
        raise FileNotFoundError(f"Model training time not found at {model_training_time_file}. Please run `python get_running_times.py` before analysis.")
    log.info(f"Loaded model training time as {model_training_time} seconds")

    # Load the cost computation time estimates
    cost_time_file = results_dir/"cost_time.txt"
    if cost_time_file.exists():
        with cost_time_file.open('r') as fp:
            cost_time = float(fp.read())
    else:
        raise FileNotFoundError(f"Model training time not found at {cost_time_file}. Please run `python get_running_times.py` before analysis.")
    log.info(f"Loaded cost computation time as {cost_time} seconds")
    # Perform offline optimization
    algorithms = []
    baselines = []
    for algorithm in cfg.algorithms:
        algorithms.append(import_module(f"algorithms.{algorithm}"))
    
    for baseline in cfg.baselines:
        baselines.append(import_module(f"baselines.{baseline}"))
    
    incremental_baselines = []
    if "incremental" in cfg.model :
        incremental_baselines.append(import_module("baselines.incremental"))
        incremental_baselines.append(import_module("baselines.adwin_incremental"))
    if "adaptive_incremental" in cfg.model:
        incremental_baselines.append(import_module("baselines.adaptive_incremental"))
    
    start = time()
    offline_df = offline_optimize(costs["offline_C"],algorithms,max_retrain_cost,use_ray=cfg.use_ray)
    # find the retraining costs for evaluation
    retrain_costs = offline_df.retrain_cost.unique()

    offline_baseline_df = gather_baselines(offline_models,offline_datas,offline_queries,baselines,retrain_costs)
    if len(incremental_baselines)>0:
        offline_incremental_baselines_df = gather_incremental_baselines(cfg,offline_datas,offline_queries,incremental_baselines,retrain_costs)
    elapsed = time() - start 
    log.info(f"Offline Retrain Cost analysis took: {elapsed:g}sec")
    
    # Perform Online Evaluation
    start = time()
    online_df = online_execute(costs["online_C"],algorithms,offline_df,use_ray=cfg.use_ray)
    online_baseline_df = gather_baselines(online_models,online_datas,online_queries,baselines,retrain_costs)
    if len(incremental_baselines)>0:
        online_incremental_baselines_df = gather_incremental_baselines(cfg,online_datas,online_queries,incremental_baselines,retrain_costs)
    elapsed = time() - start 
    log.info(f"Online Retrain Cost analysis took: {elapsed:g}sec")
    
    # compute the metrics
    
    # the number of batches where decisions need to be taken
    num_offline_batches = len(offline_datas)-1
    offline_time_kwargs = {"num_batches":num_offline_batches,"model_training_time":model_training_time,"cost_time":cost_time}
    for _df in [offline_df,offline_baseline_df]:
        _df["strategy_cost"] = _df.apply(lambda row: _compute_strategy_cost(row.retrains,costs["offline_C"],row.retrain_cost),axis=1)
        _df["prequential_acc"] = _df.apply(lambda row: _compute_prequential_accuracy(row.retrains,costs["online_data_acc"]),axis=1)
        _df["prequential_query_acc"] = _df.apply(lambda row: _compute_prequential_accuracy(row.retrains,costs["online_query_acc"]),axis=1)
        _df["total_running_time"] = _df.apply(gather_total_time,axis=1,**offline_time_kwargs)
        _df["model_training_time"] = _df.apply(lambda row:row.num_retrains*model_training_time,axis=1)
    
    # the number of batches where decisions need to be taken
    num_online_batches = len(offline_datas)-1
    online_time_kwargs = {"num_batches":num_online_batches,"model_training_time":model_training_time,"cost_time":cost_time}
    for _df in [online_df,online_baseline_df]:
        _df["strategy_cost"] = _df.apply(lambda row: _compute_strategy_cost(row.retrains,costs["online_C"],row.retrain_cost),axis=1)
        _df["prequential_acc"] = _df.apply(lambda row: _compute_prequential_accuracy(row.retrains,costs["online_data_acc"]),axis=1)
        _df["prequential_query_acc"] = _df.apply(lambda row: _compute_prequential_accuracy(row.retrains,costs["online_query_acc"]),axis=1)
        _df["total_running_time"] = _df.apply(gather_total_time,axis=1,**online_time_kwargs)
        _df["model_training_time"] = _df.apply(lambda row:row.num_retrains*model_training_time,axis=1)
    
    # join the dataframes together
    algorithm_df = join_dfs(offline_df,online_df)
    baselines_df = join_dfs(offline_baseline_df,online_baseline_df)
    if len(incremental_baselines)>0:
        incremental_baselines_df = join_dfs(offline_incremental_baselines_df,online_incremental_baselines_df)
        baselines_df = pd.concat([baselines_df,incremental_baselines_df]).reset_index(drop=True)
    # join all the results together
    df = pd.concat([algorithm_df,baselines_df]).reset_index(drop=True)

    log.info(f"Results stored in {seed_results_dir}")
    if cfg.save_df:
        save_dfs(df,seed_results_dir)
    
    log.info("Logging Configs and Logs to results directorty")
    log_dir = seed_results_dir/"analysis_logs"
    hydra_log_dir = HydraConfig.get().runtime.output_dir
    shutil.copytree(hydra_log_dir,str(log_dir),dirs_exist_ok=True)

    # print(OmegaConf.to_yaml(cfg))

if __name__ == "__main__":
    run()
