# %% [markdown]
# # Proof of Concept Scenarios

# %%
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from IPython import get_ipython
if get_ipython() is not None and __name__ == "__main__":
    notebook = True
    get_ipython().run_line_magic("load_ext", "autoreload")
    get_ipython().run_line_magic("autoreload", "2")
else:
    notebook = False
from pathlib import Path
import sys
project_root = Path(__file__).resolve().parent.parent
sys.path.append(str((project_root/"src").resolve()))

# %%
from sklearn.datasets import make_blobs, make_classification
from sklearn.linear_model import SGDClassifier
from numpy.random import default_rng
import matplotlib as mpl
from qdm import bayes_query_data_model_loss,compute_qdm_diff
from utils import plot_retrains,add_retraining_cost,_compute_strategy_cost
from algorithms import oracle
from sklearn.preprocessing import MinMaxScaler
mpl.rcParams["figure.dpi"]=100
save = False
figure_dir = project_root/"paper"/"EuroSys_2024"/"figs"/"poc"
from plotting import set_size,fullwidth,columnwidth
#%%
def annotate_axes(fig):
    for i, ax in enumerate(fig.axes):
        ax.text(0.5, 0.5, "ax%d" % (i+1), va="center", ha="center")
        ax.tick_params(labelbottom=False, labelleft=False)


def orig(n_samples=50):

    scaler = MinMaxScaler()
    # generate data
    X,y,centres = make_blobs(n_samples=n_samples,
    n_features=2,
    centers=2,
    random_state=42,
    cluster_std=3,
    return_centers=True
    )

    X = scaler.fit_transform(X)

    # put into dict for QDM computation
    orig_data = {"X_train":X,"y_train":y}

    # train model 
    model = SGDClassifier(loss="log_loss",random_state=42)
    model.fit(X,y)

    return scaler,orig_data,model,centres


def get_scen3(center,scaler,orig_model,orig_data,slopes=[2,2.5,3],intercepts=None,seed=None):
    T=3
    query_center = scaler.transform([center])[0]
    cov_mat = np.array([[0.001,0],[0,0.001]])
    # query_center = scaler.transform([[3,15]])[0]
    rng = default_rng(seed)
    _Q = rng.multivariate_normal(query_center,cov_mat,num_queries)
    _query_data = {"X_query":_Q}

    query_datas3 = {t:_query_data for t in range(0,T+1)}
    # original model parameters
    w = orig_model.coef_[0]
    b = orig_model.intercept_
    orig_intercept = -b/[w[1]]
    # all new data dictionaries starting with initial training data
    new_datas ={0:orig_data}
    # slopes for the lines
    slopes  = {t+1:s for t,s in enumerate(slopes)}
    if intercepts is None:
        intercepts = {t+1:i for t,i in enumerate([orig_intercept]*3)}
    else:
        intercepts = {t+1:i for t,i in enumerate(intercepts)}
    # dictionary of all models starting with initial model
    models = {0:orig_model}
    X,y = orig_data["X_train"],orig_data["y_train"]
    for t in range(1,T+1):
        # slope and bias of new model
        m = slopes[t]
        # keep the bias the same for sake of example
        c= intercepts[t]
        # color the same X points based on the new model
        _y = (X[:,1]<m*X[:,0]+c).astype(int)

        new_data = {"X_train":X.copy(),"y_train":_y}
        new_datas[t]=new_data
        # train models for each new data 
        _model = SGDClassifier(loss="log_loss",random_state=0).fit(X,_y)
        models[t] = _model
    
    return query_datas3,new_datas,models,slopes

def get_scen1(scaler,
              orig_model,
              orig_data,
              num_queries=5,
              seed=42,
              query_centres = np.array([
                        [-6,15],
                        [-2,10],
                        [1,5.5]])):
    T=3
    rng = default_rng(seed=seed)
    # scale queries
    query_centres = scaler.transform(query_centres)
    # get queries
    cov_mat = np.array([[0.001,0],[0,0.001]])
    query_datas1 = {}
    for t,center in enumerate(query_centres):
        _Q = rng.multivariate_normal(center,cov_mat,num_queries)
        query_datas1[t+1]={"X_query":_Q}
    # just use the queries from t=1 for t=0 as well
    query_datas1[0] = query_datas1[1]

    # dictionary of all models starting with initial model
    models = {0:orig_model}
    # all new data dictionaries starting with initial training data
    new_datas = {0:orig_data}
    # data doesn't chance so is just model data from t=0
    for (t,data) in enumerate([orig_data]*3):
        _model = SGDClassifier(loss="log",random_state=0).fit(data["X_train"],data["y_train"])
        models[t+1] = _model
        new_datas[t+1] = data
    
    return query_datas1,new_datas,models

def get_qdm(T,query_datas,new_datas,models,gamma=20):
    qdm = np.full((T+1,T+1),-np.inf)
    for i in range(T+1):
        for j in range(i,T+1):
            # model trained at timestep i
            _model = models[i]
            _model_data = new_datas[i]
            # deployed at timestep j 
            _new_data = new_datas[j]
            _query_data = query_datas[j]
            QDM_diff = compute_qdm_diff(_model,_model_data,_new_data,_query_data,gamma=2)
            qdm[i,j] = QDM_diff

    return qdm

def get_staleness(T,query_datas,new_datas,models,gamma=20):
    _staleness = np.full((T+1,T+1),-np.inf)
    for i in range(T+1):
        for j in range(i,T+1):
            # model trained at timestep i
            _model = models[i]
            _model_data = new_datas[i]
            # deployed at timestep j 
            _new_data = new_datas[j]
            _query_data = query_datas[j]
            staleness = bayes_query_data_model_loss(model=_model,data=_new_data,query_data=_query_data,gamma=gamma)
            _staleness[i,j] = staleness

    return _staleness



def plot_data(data,ax,**kwargs):
    X,y = data["X_train"],data["y_train"]
    class1 = y==0
    class1_kwargs = dict(edgecolor="purple",color="None",marker="s",label="Data (class 1)")
    class2_kwargs = dict(edgecolor="gold",color="None",marker="^",label="Data (class 0)")
    kwargs = dict(class1=class1_kwargs,class2=class2_kwargs)
    _plot_data(data,ax,**kwargs)

def _plot_data(data,ax,**kwargs):
    class1_kwargs = kwargs.pop("class1",{})
    class2_kwargs = kwargs.pop("class2",{})
    X,y = data["X_train"],data["y_train"]
    class1 = y==0
    ax.scatter(X[class1,0],X[class1,1],**class1_kwargs)
    ax.scatter(X[~class1,0],X[~class1,1],**class2_kwargs)
    ax.set_xticks([])
    ax.set_yticks([])

def plot_model(model,xmin,xmax,ymin,ymax,ax,**kwargs):
    # extract model parameters
    w = model.coef_[0]
    b = model.intercept_
    slope = -w[0]/w[1]
    intercept = -b/w[1]
    kwargs = {"color":"blue","linestyle":"--","label":"Initial Model $M_0$"}
    _plot_model(slope,intercept,xmin,xmax,ymin,ymax,ax,**kwargs)

def _plot_model(slope,intercept,xmin,xmax,ymin,ymax,ax,**kwargs):
    x = np.linspace(xmin,xmax)
    # draw model 
    line = slope*x + intercept
    ax.plot(x,line,**kwargs)
    ax.set_ylim(ymin-0.25,ymax+0.25)

def plot_concept(model,slope,ax):
    pass

def plot_queries(queries,ax):
    _Q = queries["X_query"]
    ax.scatter(_Q[:,0],_Q[:,1],edgecolor="tab:green",c="None",label="Queries",**kwargs)

def plot_poc(save_fig = False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':7,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    model_kwargs = dict(color="blue",linestyle="--",label="Initial Model $M_0$")
    T = 3
    figsize = np.array(set_size(fullwidth,subplots=(2,3)))
    # figsize[1]*=1.2
    fig = plt.figure(figsize=figsize)
    subfigs = fig.subfigures(1,2,width_ratios=[1.2,2])
    ax1 = subfigs[0].subplots(1,1)
    # subfigs[0].set_facecolor('0.85')
    # subfigs[1].set_facecolor('0.75')
    subsubfigs = subfigs[1].subfigures(2,1)
    ((ax2,ax4),(ax3,ax5)) = subfigs[1].subplots(2,2)
    # ax2,ax3 = subsubfigs[0].subplots(1,2)
    # ax4,ax5 = subsubfigs[1].subplots(1,2) 
    # annotate_axes(fig)

    scaler,orig_data,orig_model,_=orig(200)
    # plotting
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)
    plot_data(orig_data,ax=ax1)
    plot_model(orig_model,x1min,x1max,x2min,x2max,ax=ax1,**model_kwargs)
    # ax1.set_xlabel("Feature 1")
    # ax1.set_ylabel("Feature 2")
    # ax1.legend()
    ax1.set_title("$(a)~t=0$")

    retrain_cost = 1
    # subfigs[1].suptitle("(a)")
    # subsubfigs[0].suptitle("(a)")
    # subsubfigs[1].suptitle("(b)")
    queries,datas,models,slopes = get_scen3([-7,15],scaler,orig_model,orig_data)
    qdm1 = get_qdm(3,queries,datas,models)
    oracle_solution = oracle.run(qdm1,retrain_cost)["retrains"]
    im = plot_retrains(add_retraining_cost(qdm1,retrain_cost),[oracle_solution],["r"],ax=ax3)
    fig.colorbar(im,ax=ax3,shrink=1)
    ax3.set_title("$(c)$")
    ax3.set_xticks(range(0,T+1))
    ax3.set_yticks(range(0,T+1))
    ax3.set_ylabel("Batch Trained ($t^\prime$)")
    ax3.set_xlabel("Batch Evaluated ($t$)")
    
    plot_data(datas[2],ax=ax2)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax2,**model_kwargs)
    plot_queries(queries[2],ax=ax2)
    ax2.set_title("$t=3$\n$(b)$")
    
    queries,datas,models,slopes = get_scen3([3,15],scaler,orig_model,orig_data,num_queries=20)
    qdm2 = get_qdm(3,queries,datas,models)
    oracle_solution = oracle.run(qdm2,retrain_cost)["retrains"]
    im = plot_retrains(add_retraining_cost(qdm2,retrain_cost),[oracle_solution],["r"],ax=ax5)
    fig.colorbar(im,ax=ax5,shrink=1)
    ax5.set_title("$(e)$")
    ax5.set_xticks(range(0,T+1))
    ax5.set_yticks(range(0,T+1))
    ax5.set_ylabel("Batch Trained ($t^\prime$)")
    ax5.set_xlabel("Batch Evaluated ($t$)")
    
    plot_data(datas[2],ax=ax4)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax4)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax4,**model_kwargs)
    plot_queries(queries[2],ax=ax4)
    ax4.set_title("$t=3$\n$(d)$")
    # fig.subplots_adjust(left=0)
    # fig.tight_layout()
    subfigs[0].subplots_adjust(wspace=0.0,hspace=0.0,left=0,right=0.9,top=0.9,bottom=0)
    subfigs[1].subplots_adjust(wspace=0.2,hspace=0.2,left=0,right=0.9,top=0.9,bottom=0.12)
    # subsubfigs[0].subplots_adjust(wspace=0.0,hspace=0,left=0,right=1)
    # subsubfigs[1].subplots_adjust(wspace=0.0,hspace=0,left=0,right=1)
    ax2.legend(bbox_to_anchor=(0,0),bbox_transform=fig.transFigure,ncol=2,loc="upper left")
    ax5.legend(["Retraining Strategy"],bbox_to_anchor=(0.7,0.0),bbox_transform=fig.transFigure,loc="upper center")
    if save_fig:
        plt.savefig(figure_dir/"proof_of_concept.pdf",bbox_inches="tight")
    else:
        plt.show()
    

def plot_scenario1(n_samples=200,n_queries=20,save_fig=False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':7,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    figsize = np.array(set_size(fullwidth,subplots=(3,4)))
    # figsize[1]*=1.2
    fig = plt.figure(figsize=figsize)
    subfigs = fig.subfigures(2,1,height_ratios=[1,2])
    ax1 = subfigs[0].subplots(1,4)
    (ax2,ax3) = subfigs[1].subplots(1,2)
    scaler,orig_data,orig_model,orig_centers =orig(n_samples=n_samples)
    # plotting
    
    queries,datas,models = get_scen1(scaler,orig_model,orig_data,num_queries=n_queries)
    T = 3
    gamma=1/2
    x1max,x2max = -np.inf,-np.inf
    x1min,x2min = np.inf,np.inf
    # find global min and max
    for data in datas.values():
        _x1max,_x2max = data["X_train"].max(axis=0)
        _x1min,_x2min = data["X_train"].min(axis=0)
        x1min = min(x1min,_x1min)
        x2min = min(x2min,_x2min)
        x1max = max(x1max,_x1max)
        x2max = max(x2max,_x2max)
    init_model_args= dict(linestyle="--",color="blue",label="Initial Model $M_0$")

    for t in range(0,T+1): 
        _ax = ax1[t]
        plot_data(datas[t],ax=_ax)
        # plot init model
        plot_model(models[0],x1min,x1max,x2min,x2max,ax=_ax,**init_model_args)
        if t!=0:
            plot_queries(queries[t],ax=_ax)
        _ax.set_title(f"$({chr(97+t)})~t={t}$")
        # _ax.legend()
    ax1[1].legend(bbox_to_anchor=(0.5,-0.05),bbox_transform=subfigs[0].transFigure,ncol=4,loc="upper center")

    
    retrain_cost = 0
    
    staleness = get_staleness(T,queries,datas,models,gamma=gamma)
    oracle_solution = oracle.run(staleness,retrain_cost)["retrains"]
    im = ax2.imshow(add_retraining_cost(staleness,retrain_cost))
    fig.colorbar(im,ax=ax2,shrink=1)
    ax2.set_xticks(range(0,T+1))
    ax2.set_yticks(range(0,T+1))
    ax2.set_xlabel("Batch Trained ($t^\prime$)")
    ax2.set_ylabel("Batch Evaluated ($t$)")
    ax2.set_title("(e) $\Psi$ Cost Matrix")

    qdm2 = get_qdm(T,queries,datas,models,gamma=gamma)
    im = ax3.imshow(add_retraining_cost(qdm2,retrain_cost))
    fig.colorbar(im,ax=ax3,shrink=1)
    ax3.set_xticks(range(0,T+1))
    ax3.set_yticks(range(0,T+1))
    ax3.set_xlabel("Batch Trained ($t^\prime$)")
    ax3.set_ylabel("Batch Evaluated ($t$)")
    ax3.set_title("(f) $\overline{\Psi}$ Cost Matrix")

    if save_fig:
        fig.savefig(figure_dir/"scenario2.pdf",bbox_inches="tight")
    else:
        plt.show()
    
    
# %%

def plot_mini_poc1(n_samples=50,num_queries=50,save_fig = False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    T = 3
    figsize = np.array(set_size(columnwidth,subplots=(1,3)))
    # fig = plt.figure(constrained_layout=False,figsize=figsize)
    # gs = fig.add_gridspec(nrows=2, ncols=2,width_ratios=[1,2] )#left=0.05, right=0.48, wspace=0.05)
    # ax1 = fig.add_subplot(gs[:, 0])
    # ax2 = fig.add_subplot(gs[0,1])
    # ax3 = fig.add_subplot(gs[1,1])
    figsize[1]*=2
    fig, (ax1,ax2,ax3) = plt.subplots(1,3,figsize=figsize)
    scaler,orig_data,orig_model,_=orig(n_samples)
    # plotting
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)
    plot_data(orig_data,ax=ax1)
    plot_model(orig_model,x1min,x1max,x2min,x2max,ax=ax1)
    # ax1.set_xlabel("Feature 1")
    # ax1.set_ylabel("Feature 2")
    # ax1.legend()
    ax1.set_title("(a) $t=0$")

    # subfigs[1].suptitle("(a)")
    # subsubfigs[0].suptitle("(a)")
    # subsubfigs[1].suptitle("(b)")
    queries,datas,models,slopes = get_scen3([-7,12],scaler,orig_model,orig_data,num_queries=num_queries)
    
    model_kwargs = dict(color="blue",linestyle="--",label="Initial Model $M_0$")
    plot_data(datas[2],ax=ax2)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax2,**model_kwargs)
    plot_queries(queries[2],ax=ax2)
    ax2.set_title("(b) $t=1$")
    
    queries,datas,models,slopes = get_scen3([-0.5,6],scaler,orig_model,orig_data,num_queries=num_queries)
    
    plot_data(datas[2],ax=ax3)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax4)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax3)
    plot_queries(queries[2],ax=ax3)
    ax3.set_title("(c) $t=1$")

    # subfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[1].subplots_adjust(wspace=0,hspace=0)
    fig.tight_layout()
    ax2.legend(bbox_to_anchor=(0.5,0.12),bbox_transform=fig.transFigure,ncol=2,loc="upper center")
    # ax5.legend(["Retraining Strategy"],bbox_to_anchor=(0.7,0.0),bbox_transform=fig.transFigure,ncol=4)
    if save_fig:
        plt.savefig(figure_dir/"mini_proof_of_concept.pdf",bbox_inches="tight")
    else:
        plt.show()

def plot_mini_poc2(n_samples=50,num_queries=50,save_fig = False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    T = 3
    figsize = np.array(set_size(columnwidth,subplots=(1,3)))
    # fig = plt.figure(constrained_layout=False,figsize=figsize)
    # gs = fig.add_gridspec(nrows=2, ncols=2,width_ratios=[1,2] )#left=0.05, right=0.48, wspace=0.05)
    # ax1 = fig.add_subplot(gs[:, 0])
    # ax2 = fig.add_subplot(gs[0,1])
    # ax3 = fig.add_subplot(gs[1,1])
    figsize[1]*=2
    fig, (ax1,ax2,ax3) = plt.subplots(1,3,figsize=figsize)
    scaler,orig_data,orig_model,_=orig(n_samples)
    # plotting
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)

    model_kwargs = dict(color="blue",linestyle="--",label="Initial Model $M_0$")
    retrain_cost = 1
    # subfigs[1].suptitle("(a)")
    # subsubfigs[0].suptitle("(a)")
    # subsubfigs[1].suptitle("(b)")
    queries,datas,models = get_scen1(scaler,orig_model,orig_data,num_queries=num_queries)

    plot_data(datas[1],ax=ax1)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax1,**model_kwargs)
    # ax1.set_xlabel("Feature 1")
    # ax1.set_ylabel("Feature 2")
    # ax1.legend()
    plot_queries(queries[1],ax=ax1)
    ax1.set_title("(a) $t=1$")    

    plot_data(datas[1],ax=ax2)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax2,**model_kwargs)
    plot_queries(queries[2],ax=ax2)
    ax2.set_title("(b) $t=2$")
   
  
    plot_data(datas[3],ax=ax3)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax4)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax3,**model_kwargs)
    plot_queries(queries[3],ax=ax3)
    ax3.set_title("(c) $t=3$")

    # subfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[1].subplots_adjust(wspace=0,hspace=0)
    fig.tight_layout()
    ax1.legend(bbox_to_anchor=(0.5,0.12),bbox_transform=fig.transFigure,ncol=2,loc="upper center")
    # ax5.legend(["Retraining Strategy"],bbox_to_anchor=(0.7,0.0),bbox_transform=fig.transFigure,ncol=4)
    if save_fig:
        plt.savefig(figure_dir/"mini_proof_of_concept2.pdf",bbox_inches="tight")
    else:
        plt.show()
# %%

def plot_regular_poc1(n_samples=50,num_queries=50,save_fig = False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    T = 3
    figsize = np.array(set_size(fullwidth,subplots=(1,3)))
    # fig = plt.figure(constrained_layout=False,figsize=figsize)
    # gs = fig.add_gridspec(nrows=2, ncols=2,width_ratios=[1,2] )#left=0.05, right=0.48, wspace=0.05)
    # ax1 = fig.add_subplot(gs[:, 0])
    # ax2 = fig.add_subplot(gs[0,1])
    # ax3 = fig.add_subplot(gs[1,1])
    # figsize[1]*=2
    model_kwargs = dict(color="blue",linestyle="--",label="Initial Model $M_0$")
    data_kwargs = dict(alpha=0.7)
    fig, (ax1,ax2,ax3) = plt.subplots(1,3,figsize=figsize)
    scaler,orig_data,orig_model,_=orig(n_samples)
    # plotting
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)
    plot_data(orig_data,ax=ax1,**data_kwargs)
    plot_model(orig_model,x1min,x1max,x2min,x2max,ax=ax1,**model_kwargs)
    # ax1.set_xlabel("Feature 1")
    # ax1.set_ylabel("Feature 2")
    # ax1.legend()
    ax1.set_title("(a) $t=0$")

    # subfigs[1].suptitle("(a)")
    # subsubfigs[0].suptitle("(a)")
    # subsubfigs[1].suptitle("(b)")
    queries,datas,models,slopes = get_scen3([-7,13],scaler,orig_model,orig_data,num_queries=num_queries)
    
    plot_data(datas[2],ax=ax2,**data_kwargs)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax2,**model_kwargs)
    plot_queries(queries[2],ax=ax2,**data_kwargs)
    ax2.set_title("(b) $t=1$")
    
    queries,datas,models,slopes = get_scen3([-0.5,6],scaler,orig_model,orig_data,num_queries=num_queries)
    
    plot_data(datas[2],ax=ax3,**data_kwargs)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax4)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax3,**model_kwargs)
    plot_queries(queries[2],ax=ax3,**data_kwargs)
    ax3.set_title("(c) $t=1$")

    # subfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[1].subplots_adjust(wspace=0,hspace=0)
    fig.tight_layout()
    ax2.legend(bbox_to_anchor=(0.5,0.12),bbox_transform=fig.transFigure,ncol=4,loc="upper center")
    # ax5.legend(["Retraining Strategy"],bbox_to_anchor=(0.7,0.0),bbox_transform=fig.transFigure,ncol=4)
    if save_fig:
        plt.savefig(figure_dir/"regular_proof_of_concept.pdf",bbox_inches="tight")
    else:
        plt.show()
#%%
def plot_regular_poc2(n_samples=50,num_queries=50,save_fig = False):
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 7,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":7,
    "ytick.labelsize":7,
    "legend.fontsize":7,
    })
    T = 3
    figsize = np.array(set_size(fullwidth,subplots=(1,3)))
    # fig = plt.figure(constrained_layout=False,figsize=figsize)
    # gs = fig.add_gridspec(nrows=2, ncols=2,width_ratios=[1,2] )#left=0.05, right=0.48, wspace=0.05)
    # ax1 = fig.add_subplot(gs[:, 0])
    # ax2 = fig.add_subplot(gs[0,1])
    # ax3 = fig.add_subplot(gs[1,1])
    # figsize[1]*=2
    fig, (ax1,ax2,ax3) = plt.subplots(1,3,figsize=figsize)
    scaler,orig_data,orig_model,_=orig(n_samples)
    # plotting
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)

    model_kwargs = dict(color="blue",linestyle="--",label="Initial Model $M_0$")
    data_kwargs = dict(alpha=0.7)
    retrain_cost = 1
    # subfigs[1].suptitle("(a)")
    # subsubfigs[0].suptitle("(a)")
    # subsubfigs[1].suptitle("(b)")
    queries,datas,models = get_scen1(scaler,orig_model,orig_data,num_queries=num_queries,query_centres=np.array([
                        [-7,15],
                        [-2.5,11],
                        [0,6]]))

    plot_data(datas[1],ax=ax1,**data_kwargs)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax1,**model_kwargs)
    # ax1.set_xlabel("Feature 1")
    # ax1.set_ylabel("Feature 2")
    # ax1.legend()
    plot_queries(queries[1],ax=ax1,**data_kwargs)
    ax1.set_title("(a) $t=1$")    

    plot_data(datas[1],ax=ax2,**data_kwargs)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax2,**model_kwargs)
    plot_queries(queries[2],ax=ax2,**data_kwargs)
    ax2.set_title("(b) $t=2$")
   
  
    plot_data(datas[3],ax=ax3,**data_kwargs)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax4)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax3,**model_kwargs)
    plot_queries(queries[3],ax=ax3,**data_kwargs)
    ax3.set_title("(c) $t=3$")

    # subfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[0].subplots_adjust(wspace=0,hspace=0)
    # subsubfigs[1].subplots_adjust(wspace=0,hspace=0)
    fig.tight_layout()
    ax1.legend(bbox_to_anchor=(0.5,0.12),bbox_transform=fig.transFigure,ncol=4,loc="upper center")
    # ax5.legend(["Retraining Strategy"],bbox_to_anchor=(0.7,0.0),bbox_transform=fig.transFigure,ncol=4)
    if save_fig:
        plt.savefig(figure_dir/"regular_proof_of_concept2.pdf",bbox_inches="tight")
    else:
        plt.show()
#%%
if __name__ == "__main__":
    save_fig=False
    plot_mini_poc(save_fig=save_fig)
    #%%
    plot_poc(save_fig=save_fig)
    # %%
    T =3
    fig,(ax1,ax2,ax3) = plt.subplots(1,3)
    # retrain_cost = .1
    retrain_cost = 1
    scaler,orig_data,orig_model=orig()
    # plotting
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)
    plot_data(orig_data,ax=ax1)
    plot_model(orig_model,x1min,x1max,x2min,x2max,ax=ax1)
    queries,datas,models,slopes = get_scen3([-7,20],scaler,orig_model,orig_data)    
    # queries,datas,models,slopes = get_scen3([4,10],scaler,orig_model,orig_data)    
    qdm2 = get_qdm(T,queries,datas,models)
    oracle_solution = oracle.run(qdm2,retrain_cost)["retrains"]
    im = plot_retrains(add_retraining_cost(qdm2,retrain_cost),[oracle_solution],["r"],ax=ax3)
    fig.colorbar(im,ax=ax3,shrink=1)
    plot_data(datas[2],ax=ax2)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    plot_model(models[0],x1min,x1max,x2min,x2max,ax=ax2)
    plot_queries(queries[0],ax=ax2)
    #%%
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 25,
    'font.family': 'Times New Roman',
    'axes.titlesize':40,
    'figure.dpi':100,
    "xtick.labelsize":15,
    "ytick.labelsize":15,
    "legend.fontsize":40,
    })
    fig,ax = plt.subplots(2,2,figsize=(10,17))
    ax = ax.flatten()
    scaler,orig_data,orig_model=orig()
    orig_intercept = -orig_model.intercept_/orig_model.coef_[0][1]
    orig_slope = -orig_model.coef_[0][0]/orig_model.coef_[0][1]
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)
    slopes = [1.5,2,3]
    intercepts = [orig_intercept]*3
    queries,datas,models,slopes = get_scen3([-7,20],scaler,orig_model,orig_data,slopes=slopes,intercepts=intercepts)    
    concept_kwargs = {"label":"True Concept","color":"black","linestyle":"--","linewidth":3}
    init_model_kwargs = {"label":"Initial Model $M_0$","color":"blue","linestyle":"-","linewidth":3}
    class1_kwargs = dict(edgecolor="black",color="purple",marker="s",label="Data (class 1)",s=200,alpha=0.7)
    class2_kwargs = dict(edgecolor="black",color="gold",marker="^",label="Data (class 0)",s=200,alpha=0.7)
    data_kwargs = dict(class1=class1_kwargs,class2=class2_kwargs)
    for i,axes in enumerate(ax):
        _data = datas[i]
        _queries = queries[i]
        _model = models[i]
        _plot_data(_data,ax=axes,**data_kwargs)
        # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
        _plot_model(orig_slope,orig_intercept,x1min,x1max,x2min,x2max,ax=axes,**init_model_kwargs)
        # plot_model(_model,x1min,x1max,x2min,x2max,ax=axes)
        if i>0:
            _plot_model(slopes[i],intercepts[i-1],x1min,x1max,x2min,x2max,ax=axes,**concept_kwargs)
            # plot_queries(_queries,ax=axes)
        axes.set_title(f"$t={i}$")
    ax[1].legend(bbox_transform=fig.transFigure,bbox_to_anchor=(0.5,0.5),loc="center",ncol=2,frameon=False,borderpad=0,handletextpad=0.2,columnspacing=0.5,borderaxespad=0)
    fig.subplots_adjust(wspace=0.1,left=0.05,right=0.95,hspace=0.5)
    fig.savefig(project_root/"paper/AI Day poster/figs"/f"illustration.pdf",bbox_inches="tight")
    # %%
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 25,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":25,
    "ytick.labelsize":25,
    "legend.fontsize":15,
    })
    # retrain_cost = 1
    # prefix = "high_"
    retrain_cost = 0.1
    prefix = "low_"
    imshow_kwargs = {"vmin": 0, "vmax": 1,"cmap":"binary"}
    retrain_kwargs = {"linewidth":3,"markersize":10}
    kwargs = {"imshow":imshow_kwargs,"retrain":retrain_kwargs}
    values = [[0.12,0.75,0.91,0.82,0.96,0.15]]
    qdm = np.full((T+1,T+1),-np.inf)
    qdm[np.diag_indices(T+1)]=0
    qdm[np.triu_indices(T+1,1)]=values
    oracle_solution = oracle.run(qdm,retrain_cost)["retrains"]
    fig,ax = plt.subplots()
    im = plot_retrains(add_retraining_cost(qdm,retrain_cost),[oracle_solution],["r"],ax=ax,**kwargs)
    ax.legend(["Retraining Strategy"],loc="lower left")
    cbar = fig.colorbar(im,ax=ax,shrink=1)
    cbar.ax.get_yaxis().labelpad = 30
    cbar.ax.set_ylabel('cost', rotation=270)
    ax.set_xticks(range(0,T+1))
    ax.set_yticks(range(0,T+1))
    ax.set_xlabel("Batch Evaluated ($t$)")
    ax.set_ylabel("Batch Trained ($t^\prime$)")
    fig.savefig(project_root/"paper/AI Day poster/figs"/f"{prefix}retrain_cost.pdf",bbox_inches="tight")

    # %%
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 25,
    'font.family': 'Times New Roman',
    'axes.titlesize':10,
    'figure.dpi':100,
    "xtick.labelsize":25,
    "ytick.labelsize":25,
    "legend.fontsize":15,
    })
    far_values = [0.2,0.25,0.33,0.27,0.35,0.08]
    near_values = [[0.12,0.75,0.91,0.82,0.96,0.15]]
    retrain_cost = 1

    # _values = far_values
    # prefix = "far_"
    
    _values = near_values
    prefix = "near_"

    imshow_kwargs = {"vmin": 0, "vmax": 1,"cmap":"binary"}
    retrain_kwargs = {"linewidth":3,"markersize":10}
    kwargs = {"imshow":imshow_kwargs,"retrain":retrain_kwargs}
    qdm = np.full((T+1,T+1),-np.inf)
    qdm[np.diag_indices(T+1)]=0
    qdm[np.triu_indices(T+1,1)]=_values
    oracle_solution = oracle.run(qdm,retrain_cost)["retrains"]
    fig,ax = plt.subplots()
    im = plot_retrains(add_retraining_cost(qdm,retrain_cost),[oracle_solution],["r"],ax=ax,**kwargs)
    ax.legend(["Retraining Strategy"],loc="lower left")
    cbar = fig.colorbar(im,ax=ax,shrink=1)
    cbar.ax.get_yaxis().labelpad = 30
    cbar.ax.set_ylabel('cost', rotation=270)
    ax.set_xticks(range(0,T+1))
    ax.set_yticks(range(0,T+1))
    ax.set_xlabel("Batch Evaluated ($t$)")
    ax.set_ylabel("Batch Trained ($t^\prime$)")
    fig.savefig(project_root/"paper/AI Day poster/figs"/f"{prefix}queries.pdf",bbox_inches="tight")

    # %%
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 25,
    'font.family': 'Times New Roman',
    'axes.titlesize':25,
    'figure.dpi':100,
    "xtick.labelsize":15,
    "ytick.labelsize":15,
    "legend.fontsize":32,
    })
    scaler,orig_data,orig_model=orig()
    orig_intercept = -orig_model.intercept_/orig_model.coef_[0][1]
    orig_slope = -orig_model.coef_[0][0]/orig_model.coef_[0][1]
    x1max,x2max = orig_data["X_train"].max(axis=0)
    x1min,x2min = orig_data["X_train"].min(axis=0)
    slopes = [1.5,2,3]
    intercepts = [orig_intercept]*3
    queries,datas,models,slopes = get_scen3([-7,15],scaler,orig_model,orig_data,slopes=slopes,intercepts=intercepts,seed=69)    
    prefix = "far_"
    # queries,datas,models,slopes = get_scen3([2,15],scaler,orig_model,orig_data,slopes=slopes,intercepts=intercepts,seed=69)    
    # prefix="near_"
    concept_kwargs = {"label":"True Concept","color":"black","linestyle":"--","linewidth":3}
    init_model_kwargs = {"label":"Initial Model $M_0$","color":"blue","linestyle":"-","linewidth":3}
    class1_kwargs = dict(edgecolor="black",color="purple",marker="s",label="Data (class 1)",s=200,alpha=0.7)
    class2_kwargs = dict(edgecolor="black",color="gold",marker="^",label="Data (class 0)",s=200,alpha=0.7)
    query_kwargs = dict(color="tab:green",edgecolor="black",label="Queries",s=200)
    data_kwargs = dict(class1=class1_kwargs,class2=class2_kwargs)
    fig,ax = plt.subplots()
    t=3
    _data = datas[t]
    _queries = queries[t]
    _model = models[t]
    _plot_data(_data,ax=ax,**data_kwargs)
    # plot_model(models[2],x1min,x1max,x2min,x2max,ax=ax2)
    _plot_model(orig_slope,orig_intercept,x1min,x1max,x2min,x2max,ax=ax,**init_model_kwargs)
    # plot_model(_model,x1min,x1max,x2min,x2max,ax=axes)
    _plot_model(slopes[t],intercepts[t-1],x1min,x1max,x2min,x2max,ax=ax,**concept_kwargs)
    _Q = _queries["X_query"]
    ax.scatter(_Q[:,0],_Q[:,1],**query_kwargs)
    ax.set_title(f"$t={t}$")
    # ax[1].legend(bbox_transform=fig.transFigure,bbox_to_anchor=(0.5,0.5),loc="center",ncol=2,frameon=False)
    # fig.subplots_adjust(wspace=0.1,left=0.05,right=0.95,hspace=0.5)
    fig.savefig(project_root/"paper/AI Day poster/figs"/f"{prefix}queries_illustration.pdf",bbox_inches="tight")
    # %%
    plt.rcParams.update({
    'text.usetex':True,
    # 'text.latex.preamble' :r'\usepackage{libertine} \usepackage[libertine]{newtxmath}',
    'font.size': 25,
    'font.family': 'Times New Roman',
    'axes.titlesize':25,
    'figure.dpi':100,
    "xtick.labelsize":25,
    "ytick.labelsize":25,
    "legend.fontsize":15,
    })
    T = 3
    _values = [0.2,0.25,0.33,0.27,0.35,0.08]
    retrain_cost = 1
    imshow_kwargs = {"vmin": 0, "vmax": 1,"cmap":"binary"}
    retrain_kwargs = {"linewidth":3,"markersize":10}
    kwargs = {"imshow":imshow_kwargs,"retrain":retrain_kwargs}
    qdm = np.full((T+1,T+1),-np.inf)
    qdm[np.diag_indices(T+1)]=0
    qdm[np.triu_indices(T+1,1)]=_values
    fig,ax = plt.subplots()
    im = ax.imshow(add_retraining_cost(qdm,retrain_cost),**imshow_kwargs)
    # im = plot_retrains(add_retraining_cost(qdm,retrain_cost),[strategy],["r"],ax=ax,**kwargs)
    # ax.legend(["Retraining Strategy"],loc="lower left")
    cbar = fig.colorbar(im,ax=ax,shrink=1)
    cbar.ax.get_yaxis().labelpad = 30
    cbar.ax.set_ylabel('cost', rotation=270)
    ax.set_xticks(range(0,T+1))
    ax.set_yticks(range(0,T+1))
    ax.set_xlabel("Batch Evaluated ($t$)")
    ax.set_ylabel("Batch Trained ($t^\prime$)")
    fig.savefig(project_root/"paper/AI Day poster/figs"/f"cost_matrix.pdf",bbox_inches="tight")
    strategy = [0,2,3]
    fig,ax = plt.subplots()
    im = plot_retrains(add_retraining_cost(qdm,retrain_cost),[strategy],["r"],ax=ax,**kwargs)
    ax.legend(["Retraining Strategy"],loc="lower left")
    cbar = fig.colorbar(im,ax=ax,shrink=1)
    cbar.ax.get_yaxis().labelpad = 30
    cbar.ax.set_ylabel('cost', rotation=270)
    ax.set_xticks(range(0,T+1))
    ax.set_yticks(range(0,T+1))
    ax.set_xlabel("Batch Evaluated ($t$)")
    ax.set_ylabel("Batch Trained ($t^\prime$)")
    sc = _compute_strategy_cost(strategy,add_retraining_cost(qdm,retrain_cost))
    ax.set_title(f"Strategy Cost = {sc:g}")
    fig.savefig(project_root/"paper/AI Day poster/figs"/f"strategy.pdf",bbox_inches="tight")
# %%
